# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import itertools
from dataclasses import dataclass
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.internet import defer
from twisted.python import deprecate
from twisted.python import log
from twisted.python import versions

from buildbot.db import NULL
from buildbot.db import base
from buildbot.process.results import RETRY
from buildbot.util import datetime2epoch
from buildbot.util import epoch2datetime
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    import datetime


class AlreadyClaimedError(Exception):
    pass


class NotClaimedError(Exception):
    pass


@dataclass
class BuildRequestModel:
    buildrequestid: int
    buildsetid: int
    builderid: int
    buildername: str
    submitted_at: datetime.datetime
    complete_at: datetime.datetime | None = None
    complete: bool = False
    results: int | None = None
    waited_for: bool = False
    priority: int = 0

    claimed_at: datetime.datetime | None = None
    claimed_by_masterid: int | None = None

    @property
    def claimed(self) -> bool:
        return self.claimed_at is not None

    # For backward compatibility from when SsDict inherited from Dict
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'BuildRequestsConnectorComponent '
                'getBuildRequest, and getBuildRequests '
                'no longer return BuildRequest as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


@deprecate.deprecated(versions.Version("buildbot", 4, 1, 0), BuildRequestModel)
class BrDict(BuildRequestModel):
    pass


class BuildRequestsConnectorComponent(base.DBConnectorComponent):
    def _simple_sa_select_query(self):
        reqs_tbl = self.db.model.buildrequests
        claims_tbl = self.db.model.buildrequest_claims
        builder_tbl = self.db.model.builders

        from_clause = reqs_tbl.outerjoin(claims_tbl, reqs_tbl.c.id == claims_tbl.c.brid)
        from_clause = from_clause.join(builder_tbl, reqs_tbl.c.builderid == builder_tbl.c.id)

        return sa.select(
            reqs_tbl,
            claims_tbl,
            builder_tbl.c.name.label('buildername'),
        ).select_from(from_clause)

    def _saSelectQuery(self):
        reqs_tbl = self.db.model.buildrequests
        claims_tbl = self.db.model.buildrequest_claims
        bsets_tbl = self.db.model.buildsets
        builder_tbl = self.db.model.builders
        bsss_tbl = self.db.model.buildset_sourcestamps
        sstamps_tbl = self.db.model.sourcestamps

        from_clause = reqs_tbl.outerjoin(claims_tbl, reqs_tbl.c.id == claims_tbl.c.brid)
        from_clause = from_clause.join(bsets_tbl, reqs_tbl.c.buildsetid == bsets_tbl.c.id)
        from_clause = from_clause.join(bsss_tbl, bsets_tbl.c.id == bsss_tbl.c.buildsetid)
        from_clause = from_clause.join(sstamps_tbl, bsss_tbl.c.sourcestampid == sstamps_tbl.c.id)
        from_clause = from_clause.join(builder_tbl, reqs_tbl.c.builderid == builder_tbl.c.id)

        return sa.select(
            reqs_tbl,
            claims_tbl,
            sstamps_tbl.c.branch,
            sstamps_tbl.c.repository,
            sstamps_tbl.c.codebase,
            builder_tbl.c.name.label('buildername'),
        ).select_from(from_clause)

    def getBuildRequest(self, brid) -> defer.Deferred[BuildRequestModel | None]:
        def thd(conn) -> BuildRequestModel | None:
            reqs_tbl = self.db.model.buildrequests
            q = self._simple_sa_select_query()
            q = q.where(reqs_tbl.c.id == brid)
            res = conn.execute(q)
            row = res.fetchone()
            rv = None
            if row:
                rv = self._modelFromRow(row)
            res.close()
            return rv

        return self.db.pool.do(thd)

    @defer.inlineCallbacks
    def getBuildRequests(
        self,
        builderid=None,
        complete=None,
        claimed=None,
        bsid=None,
        branch=None,
        repository=None,
        resultSpec=None,
    ):
        def deduplicateBrdict(brdicts: list[BuildRequestModel]) -> list[BuildRequestModel]:
            return list(({b.buildrequestid: b for b in brdicts}).values())

        def thd(conn) -> list[BuildRequestModel]:
            reqs_tbl = self.db.model.buildrequests
            claims_tbl = self.db.model.buildrequest_claims
            sstamps_tbl = self.db.model.sourcestamps
            q = self._saSelectQuery()
            if claimed is not None:
                if isinstance(claimed, bool):
                    if not claimed:
                        q = q.where((claims_tbl.c.claimed_at == NULL) & (reqs_tbl.c.complete == 0))
                    else:
                        q = q.where(claims_tbl.c.claimed_at != NULL)
                else:
                    q = q.where(claims_tbl.c.masterid == claimed)
            if builderid is not None:
                q = q.where(reqs_tbl.c.builderid == builderid)
            if complete is not None:
                if complete:
                    q = q.where(reqs_tbl.c.complete != 0)
                else:
                    q = q.where(reqs_tbl.c.complete == 0)
            if bsid is not None:
                q = q.where(reqs_tbl.c.buildsetid == bsid)

            if branch is not None:
                q = q.where(sstamps_tbl.c.branch == branch)
            if repository is not None:
                q = q.where(sstamps_tbl.c.repository == repository)

            if resultSpec is not None:
                return deduplicateBrdict(resultSpec.thd_execute(conn, q, self._modelFromRow))

            res = conn.execute(q)

            return deduplicateBrdict([self._modelFromRow(row) for row in res.fetchall()])

        res = yield self.db.pool.do(thd)
        return res

    @defer.inlineCallbacks
    def claimBuildRequests(self, brids, claimed_at=None):
        if claimed_at is not None:
            claimed_at = datetime2epoch(claimed_at)
        else:
            claimed_at = int(self.master.reactor.seconds())

        yield self._claim_buildrequests_for_master(brids, claimed_at, self.db.master.masterid)

    @defer.inlineCallbacks
    def _claim_buildrequests_for_master(self, brids, claimed_at, masterid):
        def thd(conn):
            transaction = conn.begin()
            tbl = self.db.model.buildrequest_claims

            try:
                q = tbl.insert()
                conn.execute(
                    q,
                    [{"brid": id, "masterid": masterid, "claimed_at": claimed_at} for id in brids],
                )
            except (sa.exc.IntegrityError, sa.exc.ProgrammingError) as e:
                transaction.rollback()
                raise AlreadyClaimedError() from e

            transaction.commit()

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def unclaimBuildRequests(self, brids):
        yield self._unclaim_buildrequests_for_master(brids, self.db.master.masterid)

    @defer.inlineCallbacks
    def _unclaim_buildrequests_for_master(self, brids, masterid):
        def thd(conn):
            transaction = conn.begin()
            claims_tbl = self.db.model.buildrequest_claims

            # we'll need to batch the brids into groups of 100, so that the
            # parameter lists supported by the DBAPI aren't exhausted
            iterator = iter(brids)

            while True:
                batch = list(itertools.islice(iterator, 100))
                if not batch:
                    break  # success!

                try:
                    q = claims_tbl.delete().where(
                        claims_tbl.c.brid.in_(batch),
                        claims_tbl.c.masterid == masterid,
                    )
                    conn.execute(q)
                except Exception:
                    transaction.rollback()
                    raise

            transaction.commit()

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def completeBuildRequests(self, brids, results, complete_at=None):
        assert results != RETRY, "a buildrequest cannot be completed with a retry status!"
        if complete_at is not None:
            complete_at = datetime2epoch(complete_at)
        else:
            complete_at = int(self.master.reactor.seconds())

        def thd(conn):
            transaction = conn.begin()

            # the update here is simple, but a number of conditions are
            # attached to ensure that we do not update a row inappropriately,
            # Note that checking that the request is mine would require a
            # subquery, so for efficiency that is not checked.

            reqs_tbl = self.db.model.buildrequests

            # we'll need to batch the brids into groups of 100, so that the
            # parameter lists supported by the DBAPI aren't exhausted
            for batch in self.doBatch(brids, 100):
                q = reqs_tbl.update()
                q = q.where(reqs_tbl.c.id.in_(batch))
                q = q.where(reqs_tbl.c.complete != 1)
                res = conn.execute(q.values(complete=1, results=results, complete_at=complete_at))

                # if an incorrect number of rows were updated, then we failed.
                if res.rowcount != len(batch):
                    log.msg(
                        f"tried to complete {len(batch)} buildrequests, "
                        f"but only completed {res.rowcount}"
                    )
                    transaction.rollback()
                    raise NotClaimedError
            transaction.commit()

        yield self.db.pool.do(thd)

    def set_build_requests_priority(self, brids, priority):
        def thd(conn):
            transaction = conn.begin()

            # the update here is simple, but a number of conditions are
            # attached to ensure that we do not update a row inappropriately,
            # Note that checking that the request is mine would require a
            # subquery, so for efficiency that is not checked.

            reqs_tbl = self.db.model.buildrequests

            # we'll need to batch the brids into groups of 100, so that the
            # parameter lists supported by the DBAPI aren't exhausted
            for batch in self.doBatch(brids, 100):
                q = reqs_tbl.update()
                q = q.where(reqs_tbl.c.id.in_(batch))
                q = q.where(reqs_tbl.c.complete != 1)
                res = conn.execute(q, priority=priority)

                # if an incorrect number of rows were updated, then we failed.
                if res.rowcount != len(batch):
                    log.msg(
                        f"tried to complete {len(batch)} buildrequests, "
                        f"but only completed {res.rowcount}"
                    )
                    transaction.rollback()
                    raise NotClaimedError
            transaction.commit()

        return self.db.pool.do(thd)

    @staticmethod
    def _modelFromRow(row):
        return BuildRequestModel(
            buildrequestid=row.id,
            buildsetid=row.buildsetid,
            builderid=row.builderid,
            buildername=row.buildername,
            submitted_at=epoch2datetime(row.submitted_at),
            complete_at=epoch2datetime(row.complete_at),
            complete=bool(row.complete),
            results=row.results,
            waited_for=bool(row.waited_for),
            priority=row.priority,
            claimed_at=epoch2datetime(row.claimed_at),
            claimed_by_masterid=row.masterid,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.trial import unittest

from buildbot.schedulers.forcesched import ForceScheduler
from buildbot.test import fakedb
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import www
from buildbot.www.authz import endpointmatchers


class EndpointBase(TestReactorMixin, www.WwwTestMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield self.make_master(url='h:/a/b/')
        self.matcher = self.makeMatcher()
        self.matcher.setAuthz(self.master.authz)
        yield self.insertData()

    def makeMatcher(self):
        raise NotImplementedError()

    def assertMatch(self, match):
        self.assertTrue(match is not None)

    def assertNotMatch(self, match):
        self.assertTrue(match is None)

    @defer.inlineCallbacks
    def insertData(self):
        yield self.master.db.insert_test_data([
            fakedb.Builder(id=21, name="builder"),
            fakedb.SourceStamp(id=13, branch='secret'),
            fakedb.Master(id=1),
            fakedb.Worker(id=2, name='worker'),
            fakedb.Build(id=15, buildrequestid=16, masterid=1, workerid=2, builderid=21),
            fakedb.BuildRequest(id=16, buildsetid=17, builderid=21),
            fakedb.Buildset(id=17),
            fakedb.BuildsetSourceStamp(id=20, buildsetid=17, sourcestampid=13),
        ])


class ValidEndpointMixin:
    @defer.inlineCallbacks
    def test_invalidPath(self):
        ret = yield self.matcher.match(("foo", "bar"))
        self.assertNotMatch(ret)


class AnyEndpointMatcher(EndpointBase):
    def makeMatcher(self):
        return endpointmatchers.AnyEndpointMatcher(role="foo")

    @defer.inlineCallbacks
    def test_nominal(self):
        ret = yield self.matcher.match(("foo", "bar"))
        self.assertMatch(ret)


class AnyControlEndpointMatcher(EndpointBase):
    def makeMatcher(self):
        return endpointmatchers.AnyControlEndpointMatcher(role="foo")

    @defer.inlineCallbacks
    def test_default_action(self):
        ret = yield self.matcher.match(("foo", "bar"))
        self.assertMatch(ret)

    @defer.inlineCallbacks
    def test_get(self):
        ret = yield self.matcher.match(("foo", "bar"), action="GET")
        self.assertNotMatch(ret)

    @defer.inlineCallbacks
    def test_other_action(self):
        ret = yield self.matcher.match(("foo", "bar"), action="foo")
        self.assertMatch(ret)


class ViewBuildsEndpointMatcherBranch(EndpointBase, ValidEndpointMixin):
    def makeMatcher(self):
        return endpointmatchers.ViewBuildsEndpointMatcher(branch="secret", role="agent")

    @defer.inlineCallbacks
    def test_build(self):
        ret = yield self.matcher.match(("builds", "15"))
        self.assertMatch(ret)

    test_build.skip = "ViewBuildsEndpointMatcher is not implemented yet"  # type: ignore[attr-defined]


class StopBuildEndpointMatcherBranch(EndpointBase, ValidEndpointMixin):
    def makeMatcher(self):
        return endpointmatchers.StopBuildEndpointMatcher(builder="builder", role="owner")

    @defer.inlineCallbacks
    def test_build(self):
        ret = yield self.matcher.match(("builds", "15"), "stop")
        self.assertMatch(ret)

    @defer.inlineCallbacks
    def test_build_no_match(self):
        self.matcher.builder = "foo"
        ret = yield self.matcher.match(("builds", "15"), "stop")
        self.assertNotMatch(ret)

    @defer.inlineCallbacks
    def test_build_no_builder(self):
        self.matcher.builder = None
        ret = yield self.matcher.match(("builds", "15"), "stop")
        self.assertMatch(ret)


class ForceBuildEndpointMatcherBranch(EndpointBase, ValidEndpointMixin):
    def makeMatcher(self):
        return endpointmatchers.ForceBuildEndpointMatcher(builder="builder", role="owner")

    @defer.inlineCallbacks
    def insertData(self):
        yield super().insertData()
        self.master.allSchedulers = lambda: [
            ForceScheduler(name="sched1", builderNames=["builder"])
        ]

    @defer.inlineCallbacks
    def test_build(self):
        ret = yield self.matcher.match(("builds", "15"), "stop")
        self.assertNotMatch(ret)

    @defer.inlineCallbacks
    def test_forcesched(self):
        ret = yield self.matcher.match(("forceschedulers", "sched1"), "force")
        self.assertMatch(ret)

    @defer.inlineCallbacks
    def test_noforcesched(self):
        ret = yield self.matcher.match(("forceschedulers", "sched2"), "force")
        self.assertNotMatch(ret)

    @defer.inlineCallbacks
    def test_forcesched_builder_no_match(self):
        self.matcher.builder = "foo"
        ret = yield self.matcher.match(("forceschedulers", "sched1"), "force")
        self.assertNotMatch(ret)

    @defer.inlineCallbacks
    def test_forcesched_nobuilder(self):
        self.matcher.builder = None
        ret = yield self.matcher.match(("forceschedulers", "sched1"), "force")
        self.assertMatch(ret)


class EnableSchedulerEndpointMatcher(EndpointBase, ValidEndpointMixin):
    def makeMatcher(self):
        return endpointmatchers.EnableSchedulerEndpointMatcher(role="agent")

    @defer.inlineCallbacks
    def test_build(self):
        ret = yield self.matcher.match(("builds", "15"), "stop")
        self.assertNotMatch(ret)

    @defer.inlineCallbacks
    def test_scheduler_enable(self):
        ret = yield self.matcher.match(("schedulers", "15"), "enable")
        self.assertMatch(ret)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.trial import unittest

from buildbot_worker.util import lineboundaries

if TYPE_CHECKING:
    from buildbot_worker.util.lineboundaries import LineInfo


def join_line_info(info1: LineInfo, info2: LineInfo) -> LineInfo:
    len_text1 = len(info1[0])
    return (
        info1[0] + info2[0],
        info1[1] + [(len_text1 + index) for index in info2[1]],
        info1[2] + info2[2],
    )


class LBF(unittest.TestCase):
    def setUp(self) -> None:
        newline_re = r'(\r\n|\r(?=.)|\033\[u|\033\[[0-9]+;[0-9]+[Hf]|\033\[2J|\x08+)'
        self.lbf = lineboundaries.LineBoundaryFinder(20, newline_re)

    def test_empty_line(self) -> None:
        self.assertEqual(self.lbf.append('1234', 1.0), None)
        self.assertEqual(self.lbf.append('\n', 2.0), ('1234\n', [4], [1.0]))
        self.assertEqual(self.lbf.append('5678\n', 3.0), ('5678\n', [4], [3.0]))
        self.assertEqual(self.lbf.flush(), None)

    def test_already_terminated(self) -> None:
        self.assertEqual(self.lbf.append('abcd\ndefg\n', 1.0), ('abcd\ndefg\n', [4, 9], [1.0, 1.0]))
        self.assertEqual(self.lbf.append('xyz\n', 2.0), ('xyz\n', [3], [2.0]))
        self.assertEqual(self.lbf.flush(), None)

    def test_partial_line(self) -> None:
        self.assertEqual(self.lbf.append('hello\nworld', 1.0), ('hello\n', [5], [1.0]))
        self.assertEqual(self.lbf.flush(), ('world\n', [5], [1.0]))

    def test_empty_appends(self) -> None:
        self.assertEqual(self.lbf.append('hello ', 1.0), None)
        self.assertEqual(self.lbf.append('', 2.0), None)
        self.assertEqual(self.lbf.append('world\n', 3.0), ('hello world\n', [11], [1.0]))
        self.assertEqual(self.lbf.append('', 1.0), None)

    def test_embedded_newlines(self) -> None:
        self.assertEqual(self.lbf.append('hello, ', 1.0), None)
        self.assertEqual(self.lbf.append('cruel\nworld', 2.0), ('hello, cruel\n', [12], [1.0]))
        self.assertEqual(self.lbf.flush(), ('world\n', [5], [2.0]))

    def test_windows_newlines_folded(self) -> None:
        r"Windows' \r\n is treated as and converted to a newline"
        self.assertEqual(self.lbf.append('hello, ', 1.0), None)
        self.assertEqual(
            self.lbf.append('cruel\r\n\r\nworld', 2.0), ('hello, cruel\n\n', [12, 13], [1.0, 2.0])
        )
        self.assertEqual(self.lbf.flush(), ('world\n', [5], [2.0]))

    def test_bare_cr_folded(self) -> None:
        r"a bare \r is treated as and converted to a newline"
        self.assertEqual(
            self.lbf.append('1%\r5%\r15%\r100%\nfinished', 1.0),
            ('1%\n5%\n15%\n100%\n', [2, 5, 9, 14], [1.0, 1.0, 1.0, 1.0]),
        )
        self.assertEqual(self.lbf.flush(), ('finished\n', [8], [1.0]))

    def test_backspace_folded(self) -> None:
        r"a lot of \b is treated as and converted to a newline"
        self.lbf.append('1%\b\b5%\b\b15%\b\b\b100%\nfinished', 1.0)
        self.assertEqual(self.lbf.flush(), ('finished\n', [8], [1.0]))

    def test_mixed_consecutive_newlines(self) -> None:
        r"mixing newline styles back-to-back doesn't collapse them"
        self.assertEqual(self.lbf.append('1\r\n\n\r', 1.0), ('1\n\n', [1, 2], [1.0, 1.0]))
        self.assertEqual(self.lbf.append('2\n\r\n', 2.0), ('\n2\n\n', [0, 2, 3], [1.0, 2.0, 2.0]))

    def test_split_newlines(self) -> None:
        r"multi-character newlines, split across chunks, are converted"
        input = 'a\nb\r\nc\rd\n\re'

        for splitpoint in range(1, len(input) - 1):
            a = input[:splitpoint]
            b = input[splitpoint:]
            lines_info = []
            lines_info.append(self.lbf.append(a, 2.0))
            lines_info.append(self.lbf.append(b, 2.0))
            lines_info.append(self.lbf.flush())

            lines_info = [e for e in lines_info if e is not None]

            joined_line_info = lines_info[0]
            assert joined_line_info is not None
            for line_info in lines_info[1:]:
                assert line_info is not None
                joined_line_info = join_line_info(joined_line_info, line_info)
            self.assertEqual(
                joined_line_info,
                ('a\nb\nc\nd\n\ne\n', [1, 3, 5, 7, 8, 10], [2.0, 2.0, 2.0, 2.0, 2.0, 2.0]),
            )

    def test_split_terminal_control(self) -> None:
        """terminal control characters are converted"""
        self.assertEqual(self.lbf.append('1234\033[u4321', 1.0), ('1234\n', [4], [1.0]))
        self.assertEqual(self.lbf.flush(), ('4321\n', [4], [1.0]))
        self.assertEqual(self.lbf.append('1234\033[1;2H4321', 2.0), ('1234\n', [4], [2.0]))
        self.assertEqual(self.lbf.flush(), ('4321\n', [4], [2.0]))
        self.assertEqual(self.lbf.append('1234\033[1;2f4321', 3.0), ('1234\n', [4], [3.0]))
        self.assertEqual(self.lbf.flush(), ('4321\n', [4], [3.0]))

    def test_long_lines(self) -> None:
        """long lines are split"""
        self.assertEqual(self.lbf.append('123456789012', 1.0), None)
        self.assertEqual(
            self.lbf.append('123456789012', 2.0), ('1234567890121234567\n', [19], [1.0])
        )
        self.assertEqual(
            self.lbf.append('123456789012345', 3.0), ('8901212345678901234\n', [19], [2.0])
        )
        self.assertEqual(self.lbf.append('123456789012', 4.0), None)
        self.assertEqual(self.lbf.flush(), ('5123456789012\n', [13], [3.0]))

    def test_empty_flush(self) -> None:
        self.assertEqual(self.lbf.flush(), None)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import changesources
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin


def changeSourceKey(changeSource: changesources.ChangeSourceModel):
    return changeSource.id


class Tests(TestReactorMixin, unittest.TestCase):
    # test data

    cs42 = fakedb.ChangeSource(id=42, name='cool_source')
    cs87 = fakedb.ChangeSource(id=87, name='lame_source')

    master13 = fakedb.Master(id=13, name='m1', active=1)
    cs42master13 = fakedb.ChangeSourceMaster(changesourceid=42, masterid=13)

    master14 = fakedb.Master(id=14, name='m2', active=0)
    cs87master14 = fakedb.ChangeSourceMaster(changesourceid=87, masterid=14)

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    @defer.inlineCallbacks
    def test_findChangeSourceId_new(self):
        """findChangeSourceId for a new changesource creates it"""
        id = yield self.db.changesources.findChangeSourceId('csname')
        cs = yield self.db.changesources.getChangeSource(id)
        self.assertEqual(cs.name, 'csname')

    @defer.inlineCallbacks
    def test_findChangeSourceId_existing(self):
        """findChangeSourceId gives the same answer for the same inputs"""
        id1 = yield self.db.changesources.findChangeSourceId('csname')
        id2 = yield self.db.changesources.findChangeSourceId('csname')
        self.assertEqual(id1, id2)

    @defer.inlineCallbacks
    def test_setChangeSourceMaster_fresh(self):
        """setChangeSourceMaster with a good pair"""
        yield self.db.insert_test_data([self.cs42, self.master13])
        yield self.db.changesources.setChangeSourceMaster(42, 13)
        cs = yield self.db.changesources.getChangeSource(42)
        self.assertEqual(cs.masterid, 13)

    @defer.inlineCallbacks
    def test_setChangeSourceMaster_inactive_but_linked(self):
        """Inactive changesource but already claimed by an active master"""
        yield self.db.insert_test_data([
            self.cs87,
            self.master13,
            self.master14,
            self.cs87master14,
        ])
        with self.assertRaises(changesources.ChangeSourceAlreadyClaimedError):
            yield self.db.changesources.setChangeSourceMaster(87, 13)

    @defer.inlineCallbacks
    def test_setChangeSourceMaster_active(self):
        """Active changesource already claimed by an active master"""
        yield self.db.insert_test_data([
            self.cs42,
            self.master13,
            self.cs42master13,
        ])
        with self.assertRaises(changesources.ChangeSourceAlreadyClaimedError):
            yield self.db.changesources.setChangeSourceMaster(42, 14)

    @defer.inlineCallbacks
    def test_setChangeSourceMaster_None(self):
        """A 'None' master disconnects the changesource"""
        yield self.db.insert_test_data([
            self.cs87,
            self.master14,
            self.cs87master14,
        ])
        yield self.db.changesources.setChangeSourceMaster(87, None)
        cs = yield self.db.changesources.getChangeSource(87)
        self.assertEqual(cs.masterid, None)

    @defer.inlineCallbacks
    def test_setChangeSourceMaster_None_unowned(self):
        """A 'None' master for a disconnected changesource"""
        yield self.db.insert_test_data([self.cs87])
        yield self.db.changesources.setChangeSourceMaster(87, None)
        cs = yield self.db.changesources.getChangeSource(87)
        self.assertEqual(cs.masterid, None)

    @defer.inlineCallbacks
    def test_getChangeSource(self):
        """getChangeSource for a changesource that exists"""
        yield self.db.insert_test_data([self.cs87])
        cs = yield self.db.changesources.getChangeSource(87)
        self.assertIsInstance(cs, changesources.ChangeSourceModel)
        self.assertEqual(cs, changesources.ChangeSourceModel(id=87, name='lame_source'))

    @defer.inlineCallbacks
    def test_getChangeSource_missing(self):
        """getChangeSource for a changesource that doesn't exist"""
        cs = yield self.db.changesources.getChangeSource(87)
        self.assertEqual(cs, None)

    @defer.inlineCallbacks
    def test_getChangeSource_active(self):
        """getChangeSource for a changesource that exists and is active"""
        yield self.db.insert_test_data([self.cs42, self.master13, self.cs42master13])
        cs = yield self.db.changesources.getChangeSource(42)
        self.assertIsInstance(cs, changesources.ChangeSourceModel)
        self.assertEqual(
            cs, changesources.ChangeSourceModel(id=42, name='cool_source', masterid=13)
        )

    @defer.inlineCallbacks
    def test_getChangeSource_inactive_but_linked(self):
        """getChangeSource for a changesource that is assigned but is inactive"""
        yield self.db.insert_test_data([self.cs87, self.master14, self.cs87master14])
        cs = yield self.db.changesources.getChangeSource(87)
        self.assertIsInstance(cs, changesources.ChangeSourceModel)
        self.assertEqual(
            cs, changesources.ChangeSourceModel(id=87, name='lame_source', masterid=14)
        )
        # row exists, but marked inactive

    @defer.inlineCallbacks
    def test_getChangeSources(self):
        """getChangeSources returns all changesources"""
        yield self.db.insert_test_data([
            self.cs42,
            self.master13,
            self.cs42master13,
            self.cs87,
        ])
        cslist = yield self.db.changesources.getChangeSources()

        for cs in cslist:
            self.assertIsInstance(cs, changesources.ChangeSourceModel)

        self.assertEqual(
            sorted(cslist, key=changeSourceKey),
            sorted(
                [
                    changesources.ChangeSourceModel(id=42, name='cool_source', masterid=13),
                    changesources.ChangeSourceModel(id=87, name='lame_source', masterid=None),
                ],
                key=changeSourceKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getChangeSources_masterid(self):
        """getChangeSources returns all changesources for a given master"""
        yield self.db.insert_test_data([
            self.cs42,
            self.master13,
            self.cs42master13,
            self.cs87,
        ])
        cslist = yield self.db.changesources.getChangeSources(masterid=13)

        for cs in cslist:
            self.assertIsInstance(cs, changesources.ChangeSourceModel)

        self.assertEqual(
            sorted(cslist, key=changeSourceKey),
            sorted(
                [
                    changesources.ChangeSourceModel(id=42, name='cool_source', masterid=13),
                ],
                key=changeSourceKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getChangeSources_active(self):
        """getChangeSources for (active changesources, all masters)"""
        yield self.db.insert_test_data([self.cs42, self.master13, self.cs42master13, self.cs87])
        cslist = yield self.db.changesources.getChangeSources(active=True)

        for cs in cslist:
            self.assertIsInstance(cs, changesources.ChangeSourceModel)

        self.assertEqual(
            sorted(cslist),
            sorted([
                changesources.ChangeSourceModel(id=42, name='cool_source', masterid=13),
            ]),
        )

    @defer.inlineCallbacks
    def test_getChangeSources_active_masterid(self):
        """getChangeSources returns (active changesources, given masters)"""
        yield self.db.insert_test_data([self.cs42, self.master13, self.cs42master13, self.cs87])
        cslist = yield self.db.changesources.getChangeSources(active=True, masterid=13)

        for cs in cslist:
            self.assertIsInstance(cs, changesources.ChangeSourceModel)

        self.assertEqual(
            sorted(cslist),
            sorted([
                changesources.ChangeSourceModel(id=42, name='cool_source', masterid=13),
            ]),
        )

        cslist = yield self.db.changesources.getChangeSources(active=True, masterid=14)

        for cs in cslist:
            self.assertIsInstance(cs, changesources.ChangeSourceModel)

        self.assertEqual(sorted(cslist), [])

    @defer.inlineCallbacks
    def test_getChangeSources_inactive(self):
        """getChangeSources returns (inactive changesources, all masters)"""
        yield self.db.insert_test_data([self.cs42, self.master13, self.cs42master13, self.cs87])
        cslist = yield self.db.changesources.getChangeSources(active=False)

        for cs in cslist:
            self.assertIsInstance(cs, changesources.ChangeSourceModel)

        self.assertEqual(
            sorted(cslist),
            sorted([
                changesources.ChangeSourceModel(id=87, name='lame_source'),
            ]),
        )

    @defer.inlineCallbacks
    def test_getChangeSources_inactive_masterid(self):
        """getChangeSources returns (active changesources, given masters)"""
        yield self.db.insert_test_data([self.cs42, self.master13, self.cs42master13, self.cs87])
        cslist = yield self.db.changesources.getChangeSources(active=False, masterid=13)

        for cs in cslist:
            self.assertIsInstance(cs, changesources.ChangeSourceModel)

        self.assertEqual(sorted(cslist), [])

        cslist = yield self.db.changesources.getChangeSources(active=False, masterid=14)

        for cs in cslist:
            self.assertIsInstance(cs, changesources.ChangeSourceModel)

        self.assertEqual(sorted(cslist), [])  # always returns [] by spec!



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import calendar
import datetime
from unittest import mock

import jwt
from twisted.cred import strcred
from twisted.cred.checkers import InMemoryUsernamePasswordDatabaseDontUse
from twisted.internet import defer
from twisted.trial import unittest
from twisted.web._auth.wrapper import HTTPAuthSessionWrapper
from twisted.web.server import Request

from buildbot.test.reactor import TestReactorMixin
from buildbot.test.unit.www import test_hooks_base
from buildbot.test.util import www
from buildbot.www import auth
from buildbot.www import change_hook
from buildbot.www import resource
from buildbot.www import rest
from buildbot.www import service


class FakeChannel:
    transport = None

    def isSecure(self):
        return False

    def getPeer(self):
        return None

    def getHost(self):
        return None


class NeedsReconfigResource(resource.Resource):
    needsReconfig = True
    reconfigs = 0

    def reconfigResource(self, new_config):
        NeedsReconfigResource.reconfigs += 1


class Test(TestReactorMixin, www.WwwTestMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield self.make_master(url='h:/a/b/')
        self.svc = self.master.www = service.WWWService()
        yield self.svc.setServiceParent(self.master)

    def makeConfig(self, **kwargs):
        w = {"port": None, "auth": auth.NoAuth(), "logfileName": 'l'}
        w.update(kwargs)
        new_config = mock.Mock()
        new_config.www = w
        new_config.buildbotURL = 'h:/'
        self.master.config = new_config
        return new_config

    @defer.inlineCallbacks
    def test_reconfigService_no_port(self):
        new_config = self.makeConfig()
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)

        self.assertEqual(self.svc.site, None)

    @defer.inlineCallbacks
    def test_reconfigService_reconfigResources(self):
        new_config = self.makeConfig(port=8080)
        self.patch(rest, 'RestRootResource', NeedsReconfigResource)
        NeedsReconfigResource.reconfigs = 0

        # first time, reconfigResource gets called along with setupSite
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)
        self.assertEqual(NeedsReconfigResource.reconfigs, 1)

        # and the next time, setupSite isn't called, but reconfigResource is
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)
        self.assertEqual(NeedsReconfigResource.reconfigs, 2)

    @defer.inlineCallbacks
    def test_reconfigService_port(self):
        new_config = self.makeConfig(port=20)
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)

        self.assertNotEqual(self.svc.site, None)
        self.assertNotEqual(self.svc.port_service, None)
        self.assertEqual(self.svc.port, 20)

    @defer.inlineCallbacks
    def test_reconfigService_expiration_time(self):
        new_config = self.makeConfig(port=80, cookie_expiration_time=datetime.timedelta(minutes=1))
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)

        self.assertNotEqual(self.svc.site, None)
        self.assertNotEqual(self.svc.port_service, None)
        self.assertEqual(service.BuildbotSession.expDelay, datetime.timedelta(minutes=1))

    @defer.inlineCallbacks
    def test_reconfigService_port_changes(self):
        new_config = self.makeConfig(port=20)
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)

        newer_config = self.makeConfig(port=999)
        yield self.svc.reconfigServiceWithBuildbotConfig(newer_config)

        self.assertNotEqual(self.svc.site, None)
        self.assertNotEqual(self.svc.port_service, None)
        self.assertEqual(self.svc.port, 999)

    @defer.inlineCallbacks
    def test_reconfigService_port_changes_to_none(self):
        new_config = self.makeConfig(port=20)
        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)

        newer_config = self.makeConfig()
        yield self.svc.reconfigServiceWithBuildbotConfig(newer_config)

        # (note the site sticks around)
        self.assertEqual(self.svc.port_service, None)
        self.assertEqual(self.svc.port, None)

    def test_setupSite(self):
        self.svc.setupSite(self.makeConfig())
        site = self.svc.site

        # check that it has the right kind of resources attached to its
        # root
        root = site.resource
        req = mock.Mock()
        self.assertIsInstance(root.getChildWithDefault(b'api', req), rest.RestRootResource)

    def test_setupSiteWithProtectedHook(self):
        checker = InMemoryUsernamePasswordDatabaseDontUse()
        checker.addUser("guest", "password")

        self.svc.setupSite(
            self.makeConfig(change_hook_dialects={'base': True}, change_hook_auth=[checker])
        )
        site = self.svc.site

        # check that it has the right kind of resources attached to its
        # root
        root = site.resource
        req = mock.Mock()
        self.assertIsInstance(root.getChildWithDefault(b'change_hook', req), HTTPAuthSessionWrapper)

    @defer.inlineCallbacks
    def test_setupSiteWithHook(self):
        new_config = self.makeConfig(change_hook_dialects={'base': True})
        self.svc.setupSite(new_config)
        site = self.svc.site

        # check that it has the right kind of resources attached to its
        # root
        root = site.resource
        req = mock.Mock()
        ep = root.getChildWithDefault(b'change_hook', req)
        self.assertIsInstance(ep, change_hook.ChangeHookResource)

        # not yet configured
        self.assertEqual(ep.dialects, {})

        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)

        # now configured
        self.assertEqual(ep.dialects, {'base': True})

        rsrc = self.svc.site.resource.getChildWithDefault(b'change_hook', mock.Mock())
        path = b'/change_hook/base'
        request = test_hooks_base._prepare_request({})
        self.master.data.updates.addChange = mock.Mock()
        yield self.render_resource(rsrc, path, request=request)
        self.master.data.updates.addChange.assert_called()

    @defer.inlineCallbacks
    def test_setupSiteWithHookAndAuth(self):
        fn = self.mktemp()
        with open(fn, 'w', encoding='utf-8') as f:
            f.write("user:pass")
        new_config = self.makeConfig(
            port=8080,
            plugins={},
            change_hook_dialects={'base': True},
            change_hook_auth=[strcred.makeChecker("file:" + fn)],
        )
        self.svc.setupSite(new_config)

        yield self.svc.reconfigServiceWithBuildbotConfig(new_config)
        rsrc = self.svc.site.resource.getChildWithDefault(b'', mock.Mock())

        res = yield self.render_resource(rsrc, b'')
        self.assertIn(b'{"type": "file"}', res)

        rsrc = self.svc.site.resource.getChildWithDefault(b'change_hook', mock.Mock())
        res = yield self.render_resource(rsrc, b'/change_hook/base')
        # as UnauthorizedResource is in private namespace, we cannot use
        # assertIsInstance :-(
        self.assertIn('UnauthorizedResource', repr(res))


class TestBuildbotSite(unittest.SynchronousTestCase):
    SECRET = 'secret'

    def setUp(self):
        self.site = service.BuildbotSite(None, "logs", 0, 0)
        self.site.setSessionSecret(self.SECRET)

    def test_getSession_from_bad_jwt(self):
        """if the cookie is bad (maybe from previous version of buildbot),
        then we should raise KeyError for consumption by caller,
        and log the JWT error
        """
        with self.assertRaises(KeyError):
            self.site.getSession("xxx")
        self.flushLoggedErrors(jwt.exceptions.DecodeError)

    def test_getSession_from_correct_jwt(self):
        payload = {'user_info': {'some': 'payload'}}
        uid = jwt.encode(payload, self.SECRET, algorithm=auth.SESSION_SECRET_ALGORITHM)
        session = self.site.getSession(uid)
        self.assertEqual(session.user_info, {'some': 'payload'})

    def test_getSession_from_expired_jwt(self):
        # expired one week ago
        exp = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(weeks=1)
        exp = calendar.timegm(datetime.datetime.timetuple(exp))
        payload = {'user_info': {'some': 'payload'}, 'exp': exp}
        uid = jwt.encode(payload, self.SECRET, algorithm=auth.SESSION_SECRET_ALGORITHM)
        with self.assertRaises(KeyError):
            self.site.getSession(uid)

    def test_getSession_with_no_user_info(self):
        payload = {'foo': 'bar'}
        uid = jwt.encode(payload, self.SECRET, algorithm=auth.SESSION_SECRET_ALGORITHM)
        with self.assertRaises(KeyError):
            self.site.getSession(uid)

    def test_makeSession(self):
        session = self.site.makeSession()
        self.assertEqual(session.user_info, {'anonymous': True})

    def test_updateSession(self):
        session = self.site.makeSession()
        request = Request(FakeChannel(), False)
        request.sitepath = [b"bb"]
        session.updateSession(request)
        self.assertEqual(len(request.cookies), 1)
        _, value = request.cookies[0].split(b";")[0].split(b"=")
        decoded = jwt.decode(value, self.SECRET, algorithms=[auth.SESSION_SECRET_ALGORITHM])
        self.assertEqual(decoded['user_info'], {'anonymous': True})
        self.assertIn('exp', decoded)

    def test_absentServerHeader(self):
        request = Request(FakeChannel(), False)
        self.assertEqual(request.responseHeaders.hasHeader('Server'), False)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import base64
from dataclasses import dataclass
from typing import TYPE_CHECKING

import sqlalchemy as sa
from twisted.internet import defer
from twisted.python import deprecate
from twisted.python import log
from twisted.python import versions

from buildbot.db import base
from buildbot.util import bytes2unicode
from buildbot.util import epoch2datetime
from buildbot.util import unicode2bytes
from buildbot.util.sautils import hash_columns
from buildbot.warnings import warn_deprecated

if TYPE_CHECKING:
    import datetime


@dataclass
class PatchModel:
    patchid: int
    body: bytes
    level: int
    author: str
    comment: str
    subdir: str | None = None


@dataclass
class SourceStampModel:
    ssid: int
    branch: str | None
    revision: str | None
    repository: str
    created_at: datetime.datetime
    codebase: str = ''
    project: str = ''

    patch: PatchModel | None = None

    # For backward compatibility from when SsDict inherited from Dict
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'SourceStampsConnectorComponent '
                'getSourceStamp, get_sourcestamps_for_buildset, '
                'getSourceStampsForBuild, and getSourceStamps'
                'no longer return SourceStamp as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        # moved to PatchModel object
        patch_key = {
            'patchid': 'patchid',
            'patch_body': 'body',
            'patch_level': 'level',
            'patch_author': 'author',
            'patch_comment': 'comment',
            'patch_subdir': 'subdir',
        }.get(key)
        if patch_key is not None:
            if self.patch is None:
                return None

            return getattr(self.patch, patch_key)

        raise KeyError(key)


@deprecate.deprecated(versions.Version("buildbot", 4, 1, 0), SourceStampModel)
class SsDict(SourceStampModel):
    pass


class SourceStampsConnectorComponent(base.DBConnectorComponent):
    @defer.inlineCallbacks
    def findSourceStampId(
        self,
        branch=None,
        revision=None,
        repository=None,
        project=None,
        codebase=None,
        patch_body=None,
        patch_level=None,
        patch_author=None,
        patch_comment=None,
        patch_subdir=None,
    ):
        sourcestampid, _ = yield self.findOrCreateId(
            branch,
            revision,
            repository,
            project,
            codebase,
            patch_body,
            patch_level,
            patch_author,
            patch_comment,
            patch_subdir,
        )
        return sourcestampid

    @defer.inlineCallbacks
    def findOrCreateId(
        self,
        branch=None,
        revision=None,
        repository=None,
        project=None,
        codebase=None,
        patch_body=None,
        patch_level=None,
        patch_author=None,
        patch_comment=None,
        patch_subdir=None,
    ):
        tbl = self.db.model.sourcestamps

        assert codebase is not None, "codebase cannot be None"
        assert project is not None, "project cannot be None"
        assert repository is not None, "repository cannot be None"
        self.checkLength(tbl.c.branch, branch)
        self.checkLength(tbl.c.revision, revision)
        self.checkLength(tbl.c.repository, repository)
        self.checkLength(tbl.c.project, project)

        # get a patchid, if we have a patch
        def thd(conn):
            patchid = None
            if patch_body:
                patch_body_bytes = unicode2bytes(patch_body)
                patch_base64_bytes = base64.b64encode(patch_body_bytes)
                ins = self.db.model.patches.insert()
                r = conn.execute(
                    ins,
                    {
                        "patchlevel": patch_level,
                        "patch_base64": bytes2unicode(patch_base64_bytes),
                        "patch_author": patch_author,
                        "patch_comment": patch_comment,
                        "subdir": patch_subdir,
                    },
                )
                conn.commit()
                patchid = r.inserted_primary_key[0]
            return patchid

        patchid = yield self.db.pool.do(thd)

        ss_hash = hash_columns(branch, revision, repository, project, codebase, patchid)
        sourcestampid, found = yield self.findOrCreateSomethingId(
            tbl=tbl,
            whereclause=tbl.c.ss_hash == ss_hash,
            insert_values={
                'branch': branch,
                'revision': revision,
                'repository': repository,
                'codebase': codebase,
                'project': project,
                'patchid': patchid,
                'ss_hash': ss_hash,
                'created_at': int(self.master.reactor.seconds()),
            },
        )
        return sourcestampid, found

    # returns a Deferred that returns a value
    @base.cached("ssdicts")
    def getSourceStamp(self, ssid) -> defer.Deferred[SourceStampModel | None]:
        def thd(conn) -> SourceStampModel | None:
            tbl = self.db.model.sourcestamps
            q = tbl.select().where(tbl.c.id == ssid)
            res = conn.execute(q)
            row = res.fetchone()
            if not row:
                return None
            model = self._rowToModel_thd(conn, row)
            res.close()
            return model

        return self.db.pool.do(thd)

    # returns a Deferred that returns a value
    def get_sourcestamps_for_buildset(self, buildsetid) -> defer.Deferred[list[SourceStampModel]]:
        def thd(conn) -> list[SourceStampModel]:
            bsets_tbl = self.db.model.buildsets
            bsss_tbl = self.db.model.buildset_sourcestamps
            sstamps_tbl = self.db.model.sourcestamps

            from_clause = bsets_tbl.join(bsss_tbl, bsets_tbl.c.id == bsss_tbl.c.buildsetid).join(
                sstamps_tbl, bsss_tbl.c.sourcestampid == sstamps_tbl.c.id
            )

            q = sa.select(sstamps_tbl).select_from(from_clause).where(bsets_tbl.c.id == buildsetid)

            res = conn.execute(q)
            return [self._rowToModel_thd(conn, row) for row in res.fetchall()]

        return self.db.pool.do(thd)

    # returns a Deferred that returns a value
    def getSourceStampsForBuild(self, buildid) -> defer.Deferred[list[SourceStampModel]]:
        assert buildid > 0

        def thd(conn) -> list[SourceStampModel]:
            # Get SourceStamps for the build
            builds_tbl = self.db.model.builds
            reqs_tbl = self.db.model.buildrequests
            bsets_tbl = self.db.model.buildsets
            bsss_tbl = self.db.model.buildset_sourcestamps
            sstamps_tbl = self.db.model.sourcestamps

            from_clause = builds_tbl.join(reqs_tbl, builds_tbl.c.buildrequestid == reqs_tbl.c.id)
            from_clause = from_clause.join(bsets_tbl, reqs_tbl.c.buildsetid == bsets_tbl.c.id)
            from_clause = from_clause.join(bsss_tbl, bsets_tbl.c.id == bsss_tbl.c.buildsetid)
            from_clause = from_clause.join(
                sstamps_tbl, bsss_tbl.c.sourcestampid == sstamps_tbl.c.id
            )

            q = sa.select(sstamps_tbl).select_from(from_clause).where(builds_tbl.c.id == buildid)
            res = conn.execute(q)
            return [self._rowToModel_thd(conn, row) for row in res.fetchall()]

        return self.db.pool.do(thd)

    # returns a Deferred that returns a value
    def getSourceStamps(self) -> defer.Deferred[list[SourceStampModel]]:
        def thd(conn) -> list[SourceStampModel]:
            tbl = self.db.model.sourcestamps
            q = tbl.select()
            res = conn.execute(q)
            return [self._rowToModel_thd(conn, row) for row in res.fetchall()]

        return self.db.pool.do(thd)

    def _rowToModel_thd(self, conn, row) -> SourceStampModel:
        ssid = row.id
        model = SourceStampModel(
            ssid=ssid,
            branch=row.branch,
            revision=row.revision,
            repository=row.repository,
            codebase=row.codebase,
            project=row.project,
            created_at=epoch2datetime(row.created_at),
        )
        patchid = row.patchid

        # fetch the patch, if necessary
        if patchid is not None:
            tbl = self.db.model.patches
            q = tbl.select().where(tbl.c.id == patchid)
            res = conn.execute(q)
            row = res.fetchone()
            if row:
                model.patch = PatchModel(
                    patchid=patchid,
                    body=base64.b64decode(row.patch_base64),
                    level=row.patchlevel,
                    author=row.patch_author,
                    comment=row.patch_comment,
                    subdir=row.subdir,
                )
            else:
                log.msg(f'patchid {patchid}, referenced from ssid {ssid}, not found')
            res.close()
        return model



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from parameterized import parameterized
from twisted.trial import unittest

from buildbot.db import codebase_commits
from buildbot.db.codebase_commits import CommonCommitInfo
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util.twisted import async_to_deferred


class Tests(TestReactorMixin, unittest.TestCase):
    @async_to_deferred
    async def setUp(self) -> None:  # type: ignore[override]
        self.setup_test_reactor()
        self.master = await fakemaster.make_master(self, wantDb=True)
        await self.master.db.insert_test_data([
            fakedb.Project(id=7, name='fake_project7'),
            fakedb.Codebase(id=13, projectid=7, name='codebase1'),
            fakedb.CodebaseCommit(id=106, codebaseid=13),
            fakedb.CodebaseCommit(id=107, codebaseid=13, parent_commitid=106),
            fakedb.CodebaseCommit(id=108, codebaseid=13, parent_commitid=107),
            fakedb.CodebaseCommit(id=109, codebaseid=13, parent_commitid=108),
            fakedb.CodebaseCommit(id=110, codebaseid=13, parent_commitid=109),
            fakedb.CodebaseCommit(id=119, codebaseid=13, parent_commitid=108),
            fakedb.CodebaseCommit(id=120, codebaseid=13, parent_commitid=119),
        ])

    @async_to_deferred
    async def test_get_commit_by_revision(self) -> None:
        dbdict = await self.master.db.codebase_commits.get_commit_by_revision(
            codebaseid=13, revision='rev110'
        )

        self.assertEqual(
            dbdict,
            codebase_commits.CodebaseCommitModel(
                id=110,
                codebaseid=13,
                author='author1',
                committer='committer1',
                comments='comments1',
                when_timestamp=1234567,
                revision='rev110',
                parent_commitid=109,
            ),
        )

    @async_to_deferred
    async def test_get_commit_by_revision_does_not_exist(self) -> None:
        id = await self.master.db.codebase_commits.get_commit_by_revision(
            codebaseid=13, revision='rev_not_exist'
        )
        self.assertIsNone(id)

    @parameterized.expand([(200, 200), (110, 200), (200, 110)])
    @async_to_deferred
    async def test_get_first_common_commit_with_ranges_does_not_exist(
        self, id1: int, id2: int
    ) -> None:
        r = await self.master.db.codebase_commits.get_first_common_commit_with_ranges(id1, id2)
        self.assertIsNone(r)

    @parameterized.expand([
        ('same_commit', 110, 110, CommonCommitInfo(110, [110], [110])),
        ('same_branch1', 106, 110, CommonCommitInfo(106, [106], [106, 107, 108, 109, 110])),
        ('same_branch2', 110, 106, CommonCommitInfo(106, [106, 107, 108, 109, 110], [106])),
        ('different_branches', 110, 120, CommonCommitInfo(108, [108, 109, 110], [108, 119, 120])),
    ])
    @async_to_deferred
    async def test_get_first_common_commit_with_ranges_does_same_c(
        self, name: str, id1: int, id2: int, expected: CommonCommitInfo
    ) -> None:
        r = await self.master.db.codebase_commits.get_first_common_commit_with_ranges(id1, id2)
        self.assertEqual(r, expected)

    @async_to_deferred
    async def test_get_commits(self) -> None:
        commits = await self.master.db.codebase_commits.get_commits(codebaseid=13)
        self.assertEqual(sorted(c.id for c in commits), [106, 107, 108, 109, 110, 119, 120])

    @async_to_deferred
    async def test_get_commits_by_id(self) -> None:
        dbdicts = await self.master.db.codebase_commits.get_commits_by_id([106, 200, 107])
        self.assertEqual(
            dbdicts,
            [
                codebase_commits.CodebaseCommitModel(
                    id=106,
                    codebaseid=13,
                    author='author1',
                    committer='committer1',
                    comments='comments1',
                    when_timestamp=1234567,
                    revision='rev106',
                    parent_commitid=None,
                ),
                None,
                codebase_commits.CodebaseCommitModel(
                    id=107,
                    codebaseid=13,
                    author='author1',
                    committer='committer1',
                    comments='comments1',
                    when_timestamp=1234567,
                    revision='rev107',
                    parent_commitid=106,
                ),
            ],
        )

    @async_to_deferred
    async def test_add_commit(self) -> None:
        await self.master.db.codebase_commits.add_commit(
            codebaseid=13,
            author='new_author',
            committer='new_committer',
            comments='new_comments',
            when_timestamp=12345678,
            revision='new_revision',
            parent_commitid=110,
        )
        dbdicts = await self.master.db.codebase_commits.get_commits_by_id([121])
        self.assertEqual(
            dbdicts,
            [
                codebase_commits.CodebaseCommitModel(
                    id=121,
                    codebaseid=13,
                    author='new_author',
                    committer='new_committer',
                    comments='new_comments',
                    when_timestamp=12345678,
                    revision='new_revision',
                    parent_commitid=110,
                )
            ],
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import errno
import os
import shutil
import stat
import sys
from typing import TYPE_CHECKING

from twisted.internet import defer
from twisted.python import runtime
from twisted.trial import unittest

from buildbot_worker.commands import fs
from buildbot_worker.commands import utils
from buildbot_worker.test.fake.runprocess import Expect
from buildbot_worker.test.util.command import CommandTestMixin
from buildbot_worker.test.util.compat import skipUnlessPlatformIs

if TYPE_CHECKING:
    from typing import NoReturn

    from buildbot_worker.util.twisted import InlineCallbacksType


class TestRemoveDirectory(CommandTestMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.setUpCommand()

    @defer.inlineCallbacks
    def test_simple_real(self) -> InlineCallbacksType[None]:
        file_path = os.path.join(self.basedir, 'workdir')
        self.make_command(fs.RemoveDirectory, {'paths': [file_path]}, True)
        yield self.run_command()
        self.assertFalse(os.path.exists(os.path.abspath(file_path)))
        self.assertIn(('rc', 0), self.get_updates(), self.protocol_command.show())

    @skipUnlessPlatformIs('posix')
    @defer.inlineCallbacks
    def test_simple_posix(self) -> InlineCallbacksType[None]:
        file_path = os.path.join(self.basedir, 'remove')
        self.make_command(fs.RemoveDirectory, {'paths': [file_path]}, True)

        self.patch_runprocess(
            Expect(["rm", "-rf", file_path], self.basedir, sendRC=0, timeout=120)
            .update('header', 'headers')
            .update('stdout', '')
            .update('rc', 0)
            .exit(0)
        )

        yield self.run_command()

        self.assertEqual(self.get_updates()[-2], ('rc', 0))
        self.assertIn('elapsed', self.get_updates()[-1])

    @defer.inlineCallbacks
    def test_simple_exception_real(self) -> InlineCallbacksType[None]:
        if runtime.platformType == "posix":
            return  # we only use rmdirRecursive on windows

        def fail(dir: str) -> NoReturn:
            raise RuntimeError("oh noes")

        self.patch(utils, 'rmdirRecursive', fail)
        self.make_command(fs.RemoveDirectory, {'paths': ['workdir']}, True)
        yield self.run_command()

        self.assertIn(('rc', -1), self.get_updates(), self.protocol_command.show())

    @defer.inlineCallbacks
    def test_multiple_dirs_real(self) -> InlineCallbacksType[None]:
        paths = [os.path.join(self.basedir, 'workdir'), os.path.join(self.basedir, 'sourcedir')]
        self.make_command(fs.RemoveDirectory, {'paths': paths}, True)
        yield self.run_command()

        for path in paths:
            self.assertFalse(os.path.exists(os.path.abspath(path)))
        self.assertIn(('rc', 0), self.get_updates(), self.protocol_command.show())

    @skipUnlessPlatformIs('posix')
    @defer.inlineCallbacks
    def test_multiple_dirs_posix(self) -> InlineCallbacksType[None]:
        dir_1 = os.path.join(self.basedir, 'remove_1')
        dir_2 = os.path.join(self.basedir, 'remove_2')
        self.make_command(fs.RemoveDirectory, {'paths': [dir_1, dir_2]}, True)

        self.patch_runprocess(
            Expect(["rm", "-rf", dir_1], self.basedir, sendRC=0, timeout=120)
            .update('header', 'headers')
            .update('stdout', '')
            .update('rc', 0)
            .exit(0),
            Expect(["rm", "-rf", dir_2], self.basedir, sendRC=0, timeout=120)
            .update('header', 'headers')
            .update('stdout', '')
            .update('rc', 0)
            .exit(0),
        )

        yield self.run_command()

        self.assertEqual(self.get_updates()[-2], ('rc', 0))
        self.assertIn('elapsed', self.get_updates()[-1])

    @skipUnlessPlatformIs('posix')
    @defer.inlineCallbacks
    def test_rm_after_chmod(self) -> InlineCallbacksType[None]:
        dir = os.path.join(self.basedir, 'remove')
        self.make_command(fs.RemoveDirectory, {'paths': [dir]}, True)

        self.patch_runprocess(
            Expect(["rm", "-rf", dir], self.basedir, sendRC=0, timeout=120)
            .update('header', 'headers')
            .update('stderr', 'permission denied')
            .update('rc', 1)
            .exit(1),
            Expect(['chmod', '-Rf', 'u+rwx', dir], self.basedir, sendRC=0, timeout=120)
            .update('header', 'headers')
            .update('stdout', '')
            .update('rc', 0)
            .exit(0),
            Expect(["rm", "-rf", dir], self.basedir, sendRC=0, timeout=120)
            .update('header', 'headers')
            .update('stdout', '')
            .update('rc', 0)
            .exit(0),
        )

        yield self.run_command()

        self.assertEqual(self.get_updates()[-2], ('rc', 0))
        self.assertIn('elapsed', self.get_updates()[-1])

    @skipUnlessPlatformIs('posix')
    @defer.inlineCallbacks
    def test_rm_after_failed(self) -> InlineCallbacksType[None]:
        dir = os.path.join(self.basedir, 'remove')
        self.make_command(fs.RemoveDirectory, {'paths': [dir]}, True)

        self.patch_runprocess(
            Expect(["rm", "-rf", dir], self.basedir, sendRC=0, timeout=120)
            .update('header', 'headers')
            .update('stderr', 'permission denied')
            .update('rc', 1)
            .exit(1),
            Expect(['chmod', '-Rf', 'u+rwx', dir], self.basedir, sendRC=0, timeout=120)
            .update('header', 'headers')
            .update('stdout', '')
            .update('rc', 0)
            .exit(0),
            Expect(["rm", "-rf", dir], self.basedir, sendRC=0, timeout=120)
            .update('header', 'headers')
            .update('stdout', '')
            .update('rc', 1)
            .exit(1),
        )

        yield self.run_command()

        self.assertEqual(self.get_updates()[-2], ('rc', 1))
        self.assertIn('elapsed', self.get_updates()[-1])


class TestCopyDirectory(CommandTestMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.setUpCommand()

    @defer.inlineCallbacks
    def test_simple(self) -> InlineCallbacksType[None]:
        from_path = os.path.join(self.basedir, 'workdir')
        to_path = os.path.join(self.basedir, 'copy')
        self.make_command(fs.CopyDirectory, {'from_path': from_path, 'to_path': to_path}, True)
        yield self.run_command()

        self.assertTrue(os.path.exists(os.path.abspath(to_path)))
        self.assertIn(
            ('rc', 0),  # this may ignore a 'header' : '..', which is OK
            self.get_updates(),
            self.protocol_command.show(),
        )

    @defer.inlineCallbacks
    def test_simple_exception(self) -> InlineCallbacksType[None]:
        if runtime.platformType == "posix":
            return  # we only use rmdirRecursive on windows

        def fail(src: str, dest: str) -> NoReturn:
            raise RuntimeError("oh noes")

        self.patch(shutil, 'copytree', fail)

        from_path = os.path.join(self.basedir, 'workdir')
        to_path = os.path.join(self.basedir, 'copy')
        self.make_command(fs.CopyDirectory, {'from_path': from_path, 'to_path': to_path}, True)
        yield self.run_command()

        self.assertIn(('rc', -1), self.get_updates(), self.protocol_command.show())


class TestMakeDirectory(CommandTestMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.setUpCommand()

    @defer.inlineCallbacks
    def test_empty_paths(self) -> InlineCallbacksType[None]:
        self.make_command(fs.MakeDirectory, {'paths': []}, True)
        yield self.run_command()

        self.assertUpdates([('rc', 0)], self.protocol_command.show())

    @defer.inlineCallbacks
    def test_simple(self) -> InlineCallbacksType[None]:
        paths = [os.path.join(self.basedir, 'test_dir')]
        self.make_command(fs.MakeDirectory, {'paths': paths}, True)
        yield self.run_command()

        self.assertTrue(os.path.exists(os.path.abspath(paths[0])))
        self.assertUpdates([('rc', 0)], self.protocol_command.show())

    @defer.inlineCallbacks
    def test_two_dirs(self) -> InlineCallbacksType[None]:
        paths = [os.path.join(self.basedir, 'test-dir'), os.path.join(self.basedir, 'test-dir2')]
        self.make_command(fs.MakeDirectory, {'paths': paths}, True)
        yield self.run_command()

        for path in paths:
            self.assertTrue(os.path.exists(os.path.abspath(path)))

        self.assertUpdates([('rc', 0)], self.protocol_command.show())

    @defer.inlineCallbacks
    def test_already_exists(self) -> InlineCallbacksType[None]:
        self.make_command(
            fs.MakeDirectory, {'paths': [os.path.join(self.basedir, 'workdir')]}, True
        )
        yield self.run_command()

        self.assertUpdates([('rc', 0)], self.protocol_command.show())

    @defer.inlineCallbacks
    def test_error_existing_file(self) -> InlineCallbacksType[None]:
        paths = [os.path.join(self.basedir, 'test-file')]
        self.make_command(fs.MakeDirectory, {'paths': paths}, True)

        with open(paths[0], "w"):
            pass
        yield self.run_command()

        self.assertIn(('rc', errno.EEXIST), self.get_updates(), self.protocol_command.show())


class TestStatFile(CommandTestMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.setUpCommand()

    @defer.inlineCallbacks
    def test_non_existent(self) -> InlineCallbacksType[None]:
        path = os.path.join(self.basedir, 'no-such-file')
        self.make_command(fs.StatFile, {'path': path}, True)
        yield self.run_command()

        self.assertIn(('rc', errno.ENOENT), self.get_updates(), self.protocol_command.show())

    @defer.inlineCallbacks
    def test_directory(self) -> InlineCallbacksType[None]:
        path = os.path.join(self.basedir, 'workdir')

        self.make_command(fs.StatFile, {'path': path}, True)
        yield self.run_command()

        update = self.get_updates()[0]
        assert isinstance(update, tuple)
        self.assertTrue(stat.S_ISDIR(update[1][stat.ST_MODE]))
        self.assertIn(('rc', 0), self.get_updates(), self.protocol_command.show())

    @defer.inlineCallbacks
    def test_file(self) -> InlineCallbacksType[None]:
        path = os.path.join(self.basedir, 'test-file')

        self.make_command(fs.StatFile, {'path': path}, True)
        with open(os.path.join(self.basedir, 'test-file'), "w"):
            pass

        yield self.run_command()

        update = self.get_updates()[0]
        assert isinstance(update, tuple)
        self.assertTrue(stat.S_ISREG(update[1][stat.ST_MODE]))
        self.assertIn(('rc', 0), self.get_updates(), self.protocol_command.show())

    @defer.inlineCallbacks
    def test_file_workdir(self) -> InlineCallbacksType[None]:
        path = os.path.join(self.basedir, 'wd', 'test-file')
        self.make_command(fs.StatFile, {'path': path}, True)
        os.mkdir(os.path.join(self.basedir, 'wd'))
        with open(os.path.join(self.basedir, 'wd', 'test-file'), "w"):
            pass

        yield self.run_command()

        update = self.get_updates()[0]
        assert isinstance(update, tuple)
        self.assertTrue(stat.S_ISREG(update[1][stat.ST_MODE]))
        self.assertIn(('rc', 0), self.get_updates(), self.protocol_command.show())


class TestGlobPath(CommandTestMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.setUpCommand()

    @defer.inlineCallbacks
    def test_non_existent(self) -> InlineCallbacksType[None]:
        self.make_command(fs.GlobPath, {'path': os.path.join(self.basedir, 'no-*-file')}, True)
        yield self.run_command()

        self.assertEqual(self.get_updates()[0][1], [])
        self.assertIn(('rc', 0), self.get_updates(), self.protocol_command.show())

    @defer.inlineCallbacks
    def test_directory(self) -> InlineCallbacksType[None]:
        self.make_command(fs.GlobPath, {'path': os.path.join(self.basedir, '[wxyz]or?d*')}, True)
        yield self.run_command()

        self.assertEqual(self.get_updates()[0][1], [os.path.join(self.basedir, 'workdir')])
        self.assertIn(('rc', 0), self.get_updates(), self.protocol_command.show())

    @defer.inlineCallbacks
    def test_file(self) -> InlineCallbacksType[None]:
        self.make_command(fs.GlobPath, {'path': os.path.join(self.basedir, 't*-file')}, True)
        with open(os.path.join(self.basedir, 'test-file'), "w"):
            pass

        yield self.run_command()
        self.assertEqual(self.get_updates()[0][1], [os.path.join(self.basedir, 'test-file')])
        self.assertIn(('rc', 0), self.get_updates(), self.protocol_command.show())

    @defer.inlineCallbacks
    def test_recursive(self) -> InlineCallbacksType[None]:
        self.make_command(fs.GlobPath, {'path': os.path.join(self.basedir, '**/*.txt')}, True)
        os.makedirs(os.path.join(self.basedir, 'test/testdir'))
        with open(os.path.join(self.basedir, 'test/testdir/test.txt'), 'w'):
            pass

        yield self.run_command()
        if sys.platform == 'win32':
            filename = 'test\\testdir\\test.txt'
        else:
            filename = 'test/testdir/test.txt'

        self.assertEqual(self.get_updates()[0][1], [os.path.join(self.basedir, filename)])
        self.assertIn(('rc', 0), self.get_updates(), self.protocol_command.show())


class TestListDir(CommandTestMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.setUpCommand()

    @defer.inlineCallbacks
    def test_non_existent(self) -> InlineCallbacksType[None]:
        path = os.path.join(self.basedir, 'no-such-dir')

        self.make_command(fs.ListDir, {'path': path}, True)
        yield self.run_command()

        self.assertIn(('rc', errno.ENOENT), self.get_updates(), self.protocol_command.show())

    @defer.inlineCallbacks
    def test_dir(self) -> InlineCallbacksType[None]:
        workdir = os.path.join(self.basedir, 'workdir')
        self.make_command(fs.ListDir, {'path': workdir}, True)
        with open(os.path.join(workdir, 'file1'), "w"):
            pass
        with open(os.path.join(workdir, 'file2'), "w"):
            pass

        yield self.run_command()

        self.assertIn(('rc', 0), self.get_updates(), self.protocol_command.show())

        self.assertTrue(
            any(
                'files' in upd and sorted(upd[1]) == ['file1', 'file2']
                for upd in self.get_updates()
            ),
            self.protocol_command.show(),
        )


class TestRemoveFile(CommandTestMixin, unittest.TestCase):
    def setUp(self) -> None:
        self.setUpCommand()

    @defer.inlineCallbacks
    def test_simple(self) -> InlineCallbacksType[None]:
        workdir = os.path.join(self.basedir, 'workdir')
        file1_path = os.path.join(workdir, 'file1')
        self.make_command(fs.RemoveFile, {'path': file1_path}, True)

        with open(file1_path, "w"):
            pass
        yield self.run_command()

        self.assertFalse(os.path.exists(file1_path))
        self.assertIn(
            ('rc', 0),  # this may ignore a 'header' : '..', which is OK
            self.get_updates(),
            self.protocol_command.show(),
        )

    @defer.inlineCallbacks
    def test_simple_exception(self) -> InlineCallbacksType[None]:
        workdir = os.path.join(self.basedir, 'workdir')
        file2_path = os.path.join(workdir, 'file2')

        self.make_command(fs.RemoveFile, {'path': file2_path}, True)
        yield self.run_command()

        self.assertIn(('rc', 2), self.get_updates(), self.protocol_command.show())



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import gc
import platform
import random
import string

from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import failure
from twisted.trial import unittest

from buildbot.util import lru

# construct weakref-able objects for particular keys


def short(k):
    return set([k.upper() * 3])


def long(k):
    return set([k.upper() * 6])


class LRUCacheTest(unittest.TestCase):
    def setUp(self):
        lru.inv_failed = False
        self.lru = lru.LRUCache(short, 3)

    def tearDown(self):
        self.assertFalse(lru.inv_failed, "invariant failed; see logs")

    def check_result(self, r, exp, exp_hits=None, exp_misses=None, exp_refhits=None):
        self.assertEqual(r, exp)
        if exp_hits is not None:
            self.assertEqual(self.lru.hits, exp_hits)
        if exp_misses is not None:
            self.assertEqual(self.lru.misses, exp_misses)
        if exp_refhits is not None:
            self.assertEqual(self.lru.refhits, exp_refhits)

    def test_single_key(self):
        # just get an item
        val = self.lru.get('a')
        self.check_result(val, short('a'), 0, 1)

        # second time, it should be cached..
        self.lru.miss_fn = long
        val = self.lru.get('a')
        self.check_result(val, short('a'), 1, 1)

    def test_simple_lru_expulsion(self):
        val = self.lru.get('a')
        self.check_result(val, short('a'), 0, 1)
        val = self.lru.get('b')
        self.check_result(val, short('b'), 0, 2)
        val = self.lru.get('c')
        self.check_result(val, short('c'), 0, 3)
        val = self.lru.get('d')
        self.check_result(val, short('d'), 0, 4)
        del val
        gc.collect()

        # now try 'a' again - it should be a miss
        self.lru.miss_fn = long
        val = self.lru.get('a')
        self.check_result(val, long('a'), 0, 5)

        # ..and that expelled B, but C is still in the cache
        val = self.lru.get('c')
        self.check_result(val, short('c'), 1, 5)

    @defer.inlineCallbacks
    def test_simple_lru_expulsion_maxsize_1(self):
        self.lru = lru.LRUCache(short, 1)
        val = yield self.lru.get('a')
        self.check_result(val, short('a'), 0, 1)
        val = yield self.lru.get('a')
        self.check_result(val, short('a'), 1, 1)
        val = yield self.lru.get('b')
        self.check_result(val, short('b'), 1, 2)
        del val
        gc.collect()

        # now try 'a' again - it should be a miss
        self.lru.miss_fn = long
        val = yield self.lru.get('a')
        self.check_result(val, long('a'), 1, 3)
        del val
        gc.collect()

        # ..and that expelled B
        val = yield self.lru.get('b')
        self.check_result(val, long('b'), 1, 4)

    def test_simple_lru_expulsion_maxsize_1_null_result(self):
        # a regression test for #2011
        def miss_fn(k):
            if k == 'b':
                return None
            return short(k)

        self.lru = lru.LRUCache(miss_fn, 1)
        val = self.lru.get('a')
        self.check_result(val, short('a'), 0, 1)
        val = self.lru.get('b')
        self.check_result(val, None, 0, 2)
        del val

        # 'a' was not expelled since 'b' was None
        self.lru.miss_fn = long
        val = self.lru.get('a')
        self.check_result(val, short('a'), 1, 2)

    def test_queue_collapsing(self):
        # just to check that we're practicing with the right queue size (so
        # QUEUE_SIZE_FACTOR is 10)
        self.assertEqual(self.lru.max_queue, 30)

        for c in 'a' + 'x' * 27 + 'ab':
            res = self.lru.get(c)
        self.check_result(res, short('b'), 27, 3)

        # at this point, we should have 'x', 'a', and 'b' in the cache, and
        # 'axx..xxab' in the queue.
        self.assertEqual(len(self.lru.queue), 30)

        # This 'get' operation for an existing key should cause compaction
        res = self.lru.get('b')
        self.check_result(res, short('b'), 28, 3)

        self.assertEqual(len(self.lru.queue), 3)

        # expect a cached short('a')
        self.lru.miss_fn = long
        res = self.lru.get('a')
        self.check_result(res, short('a'), 29, 3)

    def test_all_misses(self):
        for i, c in enumerate(string.ascii_lowercase + string.ascii_uppercase):
            res = self.lru.get(c)
            self.check_result(res, short(c), 0, i + 1)

    def test_get_exception(self):
        def fail_miss_fn(k):
            raise RuntimeError("oh noes")

        self.lru.miss_fn = fail_miss_fn

        got_exc = False
        try:
            self.lru.get('abc')
        except RuntimeError:
            got_exc = True

        self.assertEqual(got_exc, True)

    def test_all_hits(self):
        res = self.lru.get('a')
        self.check_result(res, short('a'), 0, 1)

        self.lru.miss_fn = long
        for i in range(100):
            res = self.lru.get('a')
            self.check_result(res, short('a'), i + 1, 1)

    def test_weakrefs(self):
        if platform.python_implementation() == 'PyPy':
            raise unittest.SkipTest('PyPy has different behavior with regards to weakref dicts')

        res_a = self.lru.get('a')
        self.check_result(res_a, short('a'))
        # note that res_a keeps a reference to this value

        res_b = self.lru.get('b')
        self.check_result(res_b, short('b'))
        del res_b  # discard reference to b

        # blow out the cache and the queue
        self.lru.miss_fn = long
        for c in string.ascii_lowercase[2:] * 5:
            self.lru.get(c)

        # and fetch a again, expecting the cached value
        res = self.lru.get('a')
        self.check_result(res, res_a, exp_refhits=1)

        # but 'b' should give us a new value
        res = self.lru.get('b')
        self.check_result(res, long('b'), exp_refhits=1)

    def test_fuzz(self):
        chars = list(string.ascii_lowercase * 40)
        random.shuffle(chars)
        for c in chars:
            res = self.lru.get(c)
            self.check_result(res, short(c))

    def test_set_max_size(self):
        # load up the cache with three items
        for c in 'abc':
            res = self.lru.get(c)
            self.check_result(res, short(c))
        del res

        # reset the size to 1
        self.lru.set_max_size(1)
        gc.collect()

        # and then expect that 'b' is no longer in the cache
        self.lru.miss_fn = long
        res = self.lru.get('b')
        self.check_result(res, long('b'))

    def test_miss_fn_kwargs(self):
        def keep_kwargs_miss_fn(k, **kwargs):
            return set(kwargs.keys())

        self.lru.miss_fn = keep_kwargs_miss_fn

        val = self.lru.get('a', a=1, b=2)
        self.check_result(val, set(['a', 'b']), 0, 1)

    def test_miss_fn_returns_none(self):
        calls = []

        def none_miss_fn(k):
            calls.append(k)
            return None

        self.lru.miss_fn = none_miss_fn

        for _ in range(2):
            self.assertEqual(self.lru.get('a'), None)

        # check that the miss_fn was called twice
        self.assertEqual(calls, ['a', 'a'])

    def test_put(self):
        self.assertEqual(self.lru.get('p'), short('p'))
        self.lru.put('p', set(['P2P2']))
        self.assertEqual(self.lru.get('p'), set(['P2P2']))

    def test_put_nonexistent_key(self):
        self.assertEqual(self.lru.get('p'), short('p'))
        self.lru.put('q', set(['new-q']))
        self.assertEqual(self.lru.get('p'), set(['PPP']))
        self.assertEqual(self.lru.get('q'), set(['new-q']))  # updated


class AsyncLRUCacheTest(unittest.TestCase):
    def setUp(self):
        lru.inv_failed = False
        self.lru = lru.AsyncLRUCache(self.short_miss_fn, 3)

    def tearDown(self):
        self.assertFalse(lru.inv_failed, "invariant failed; see logs")

    def short_miss_fn(self, key):
        return defer.succeed(short(key))

    def long_miss_fn(self, key):
        return defer.succeed(long(key))

    def failure_miss_fn(self, key):
        return defer.succeed(None)

    def check_result(self, r, exp, exp_hits=None, exp_misses=None, exp_refhits=None):
        self.assertEqual(r, exp)
        if exp_hits is not None:
            self.assertEqual(self.lru.hits, exp_hits)
        if exp_misses is not None:
            self.assertEqual(self.lru.misses, exp_misses)
        if exp_refhits is not None:
            self.assertEqual(self.lru.refhits, exp_refhits)

    # tests

    @defer.inlineCallbacks
    def test_single_key(self):
        # just get an item
        res = yield self.lru.get('a')
        self.check_result(res, short('a'), 0, 1)

        # second time, it should be cached..
        self.lru.miss_fn = self.long_miss_fn
        res = yield self.lru.get('a')
        self.check_result(res, short('a'), 1, 1)

    @defer.inlineCallbacks
    def test_simple_lru_expulsion(self):
        res = yield self.lru.get('a')
        self.check_result(res, short('a'), 0, 1)
        res = yield self.lru.get('b')
        self.check_result(res, short('b'), 0, 2)
        res = yield self.lru.get('c')
        self.check_result(res, short('c'), 0, 3)
        res = yield self.lru.get('d')
        self.check_result(res, short('d'), 0, 4)

        gc.collect()

        # now try 'a' again - it should be a miss
        self.lru.miss_fn = self.long_miss_fn
        res = yield self.lru.get('a')
        self.check_result(res, long('a'), 0, 5)

        # ..and that expelled B, but C is still in the cache
        res = yield self.lru.get('c')
        self.check_result(res, short('c'), 1, 5)

    @defer.inlineCallbacks
    def test_simple_lru_expulsion_maxsize_1(self):
        self.lru = lru.AsyncLRUCache(self.short_miss_fn, 1)

        res = yield self.lru.get('a')
        self.check_result(res, short('a'), 0, 1)
        res = yield self.lru.get('a')
        self.check_result(res, short('a'), 1, 1)
        res = yield self.lru.get('b')
        self.check_result(res, short('b'), 1, 2)

        gc.collect()

        # now try 'a' again - it should be a miss
        self.lru.miss_fn = self.long_miss_fn
        res = yield self.lru.get('a')
        self.check_result(res, long('a'), 1, 3)

        gc.collect()

        # ..and that expelled B
        res = yield self.lru.get('b')
        self.check_result(res, long('b'), 1, 4)

    @defer.inlineCallbacks
    def test_simple_lru_expulsion_maxsize_1_null_result(self):
        # a regression test for #2011
        def miss_fn(k):
            if k == 'b':
                return defer.succeed(None)
            return defer.succeed(short(k))

        self.lru = lru.AsyncLRUCache(miss_fn, 1)

        res = yield self.lru.get('a')
        self.check_result(res, short('a'), 0, 1)
        res = yield self.lru.get('b')
        self.check_result(res, None, 0, 2)

        # 'a' was not expelled since 'b' was None
        self.lru.miss_fn = self.long_miss_fn
        res = yield self.lru.get('a')
        self.check_result(res, short('a'), 1, 2)

    @defer.inlineCallbacks
    def test_queue_collapsing(self):
        # just to check that we're practicing with the right queue size (so
        # QUEUE_SIZE_FACTOR is 10)
        self.assertEqual(self.lru.max_queue, 30)

        for c in 'a' + 'x' * 27 + 'ab':
            res = yield self.lru.get(c)
        self.check_result(res, short('b'), 27, 3)

        # at this point, we should have 'x', 'a', and 'b' in the cache, and
        # 'axx..xxab' in the queue.
        self.assertEqual(len(self.lru.queue), 30)

        # This 'get' operation for an existing key should cause compaction
        res = yield self.lru.get('b')
        self.check_result(res, short('b'), 28, 3)

        self.assertEqual(len(self.lru.queue), 3)

        # expect a cached short('a')
        self.lru.miss_fn = self.long_miss_fn
        res = yield self.lru.get('a')
        self.check_result(res, short('a'), 29, 3)

    @defer.inlineCallbacks
    def test_all_misses(self):
        for i, c in enumerate(string.ascii_lowercase + string.ascii_uppercase):
            res = yield self.lru.get(c)
            self.check_result(res, short(c), 0, i + 1)

    @defer.inlineCallbacks
    def test_get_exception(self):
        def fail_miss_fn(k):
            return defer.fail(RuntimeError("oh noes"))

        self.lru.miss_fn = fail_miss_fn

        got_exc = False
        try:
            yield self.lru.get('abc')
        except RuntimeError:
            got_exc = True

        self.assertEqual(got_exc, True)

    @defer.inlineCallbacks
    def test_all_hits(self):
        res = yield self.lru.get('a')
        self.check_result(res, short('a'), 0, 1)

        self.lru.miss_fn = self.long_miss_fn
        for i in range(100):
            res = yield self.lru.get('a')
            self.check_result(res, short('a'), i + 1, 1)

    @defer.inlineCallbacks
    def test_weakrefs(self):
        if platform.python_implementation() == 'PyPy':
            raise unittest.SkipTest('PyPy has different behavior with regards to weakref dicts')

        res_a = yield self.lru.get('a')
        self.check_result(res_a, short('a'))
        # note that res_a keeps a reference to this value

        res_b = yield self.lru.get('b')
        self.check_result(res_b, short('b'))
        del res_b  # discard reference to b

        # blow out the cache and the queue
        self.lru.miss_fn = self.long_miss_fn
        for c in string.ascii_lowercase[2:] * 5:
            yield self.lru.get(c)

        # and fetch a again, expecting the cached value
        res = yield self.lru.get('a')
        self.check_result(res, res_a, exp_refhits=1)

        # but 'b' should give us a new value
        res = yield self.lru.get('b')
        self.check_result(res, long('b'), exp_refhits=1)

    @defer.inlineCallbacks
    def test_fuzz(self):
        chars = list(string.ascii_lowercase * 40)
        random.shuffle(chars)
        for c in chars:
            res = yield self.lru.get(c)
            self.check_result(res, short(c))

    @defer.inlineCallbacks
    def test_massively_parallel(self):
        chars = list(string.ascii_lowercase * 5)

        misses = [0]

        def slow_short_miss_fn(key):
            d = defer.Deferred()
            misses[0] += 1
            reactor.callLater(0, lambda: d.callback(short(key)))
            return d

        self.lru.miss_fn = slow_short_miss_fn

        def check(c, d):
            d.addCallback(self.check_result, short(c))
            return d

        yield defer.gatherResults([check(c, self.lru.get(c)) for c in chars], consumeErrors=True)

        self.assertEqual(misses[0], 26)
        self.assertEqual(self.lru.misses, 26)
        self.assertEqual(self.lru.hits, 4 * 26)

    @defer.inlineCallbacks
    def test_slow_fetch(self):
        def slower_miss_fn(k):
            d = defer.Deferred()
            reactor.callLater(0.05, lambda: d.callback(short(k)))
            return d

        self.lru.miss_fn = slower_miss_fn

        def do_get(test_d, k):
            d = self.lru.get(k)
            d.addCallback(self.check_result, short(k))
            d.addCallbacks(test_d.callback, test_d.errback)

        ds = []
        for i in range(8):
            d = defer.Deferred()
            reactor.callLater(0.02 * i, do_get, d, 'x')
            ds.append(d)

        yield defer.gatherResults(ds, consumeErrors=True)

        self.assertEqual((self.lru.hits, self.lru.misses), (7, 1))

    @defer.inlineCallbacks
    def test_slow_failure(self):
        def slow_fail_miss_fn(k):
            d = defer.Deferred()
            reactor.callLater(0.05, lambda: d.errback(failure.Failure(RuntimeError("oh noes"))))
            return d

        self.lru.miss_fn = slow_fail_miss_fn

        @defer.inlineCallbacks
        def do_get(test_d, k):
            try:
                with self.assertRaises(RuntimeError):
                    yield self.lru.get(k)
                test_d.callback(None)
            except Exception as e:
                test_d.errback(failure.Failure(e))

        ds = []
        for i in range(8):
            d = defer.Deferred()
            reactor.callLater(0.02 * i, do_get, d, 'x')
            ds.append(d)

        for d in ds:
            yield d

    @defer.inlineCallbacks
    def test_set_max_size(self):
        # load up the cache with three items
        for c in 'abc':
            res = yield self.lru.get(c)
            self.check_result(res, short(c))

        # reset the size to 1
        self.lru.set_max_size(1)
        gc.collect()

        # and then expect that 'b' is no longer in the cache
        self.lru.miss_fn = self.long_miss_fn
        res = yield self.lru.get('b')
        self.check_result(res, long('b'))

    @defer.inlineCallbacks
    def test_miss_fn_kwargs(self):
        def keep_kwargs_miss_fn(k, **kwargs):
            return defer.succeed(set(kwargs.keys()))

        self.lru.miss_fn = keep_kwargs_miss_fn

        res = yield self.lru.get('a', a=1, b=2)
        self.check_result(res, set(['a', 'b']), 0, 1)

    @defer.inlineCallbacks
    def test_miss_fn_returns_none(self):
        calls = []

        def none_miss_fn(k):
            calls.append(k)
            return defer.succeed(None)

        self.lru.miss_fn = none_miss_fn

        for _ in range(2):
            self.assertEqual((yield self.lru.get('a')), None)

        # check that the miss_fn was called twice
        self.assertEqual(calls, ['a', 'a'])

    @defer.inlineCallbacks
    def test_put(self):
        self.assertEqual((yield self.lru.get('p')), short('p'))
        self.lru.put('p', set(['P2P2']))
        self.assertEqual((yield self.lru.get('p')), set(['P2P2']))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import os
import shutil
from typing import TYPE_CHECKING
from typing import cast

from twisted.trial.unittest import TestCase

import buildbot_worker.runprocess
from buildbot_worker.commands import utils
from buildbot_worker.test.fake import runprocess
from buildbot_worker.test.fake.protocolcommand import FakeProtocolCommand

if TYPE_CHECKING:
    from typing import Any
    from typing import Sequence
    from typing import TypeVar

    from twisted.internet.defer import Deferred

    from buildbot_worker.commands.base import Command

    CommandType = TypeVar("CommandType", bound=Command)


class CommandTestMixin:
    """
    Support for testing Command subclasses.
    """

    runprocess_patched: bool = False

    def setUpCommand(self) -> None:
        """
        Get things ready to test a Command

        Sets:
            self.basedir -- the basedir (an abs path)
            self.basedir_workdir -- os.path.join(self.basedir, 'workdir')
            self.basedir_source -- os.path.join(self.basedir, 'source')
        """
        assert isinstance(self, TestCase)

        self.basedir = os.path.abspath('basedir')
        self.basedir_workdir = os.path.join(self.basedir, 'workdir')
        self.basedir_source = os.path.join(self.basedir, 'source')

        # clean up the basedir unconditionally
        if os.path.exists(self.basedir):
            shutil.rmtree(self.basedir)

        def cleanup() -> None:
            """
            Call this from the tearDown method to clean up any leftover workdirs and do
            any additional cleanup required.
            """
            # clean up the basedir unconditionally
            if os.path.exists(self.basedir):
                shutil.rmtree(self.basedir)

            # finish up the runprocess
            if hasattr(self, 'runprocess_patched') and self.runprocess_patched:
                runprocess.FakeRunProcess.test_done()

        self.addCleanup(cleanup)

    def make_command(
        self,
        cmdclass: type[CommandType],
        args: dict[str, Any],
        makedirs: bool = False,
    ) -> CommandType:
        """
        Create a new command object, creating the necessary arguments.  The
        cmdclass argument is the Command class, and args is the args dict
        to pass to its constructor.

        This always creates the FakeProtocolCommand with a basedir (self.basedir).
        If makedirs is true, it will create the basedir and a workdir directory
        inside (named 'workdir').

        The resulting command is returned, but as a side-effect, the following
        attributes are set:

            self.cmd -- the command
        """

        # set up the workdir and basedir
        if makedirs:
            basedir_abs = os.path.abspath(os.path.join(self.basedir))
            workdir_abs = os.path.abspath(os.path.join(self.basedir, 'workdir'))
            if os.path.exists(basedir_abs):
                shutil.rmtree(basedir_abs)
            os.makedirs(workdir_abs)

        self.protocol_command = FakeProtocolCommand(basedir=self.basedir)
        self.cmd = cmdclass(self.protocol_command, 'fake-stepid', args)

        return self.cmd

    def run_command(self) -> Deferred[int]:
        """
        Run the command created by make_command.  Returns a deferred that will fire
        on success or failure.
        """
        return self.cmd.doStart()

    def get_updates(self) -> Sequence[tuple[str, Any] | str]:
        """
        Return the updates made so far
        """
        return self.protocol_command.updates

    def assertUpdates(self, updates: list[tuple[str, Any] | str], msg: str | None = None) -> None:
        """
        Asserts that self.get_updates() matches updates, ignoring elapsed time data
        """
        assert isinstance(self, TestCase)
        my_updates: list[tuple[str, Any] | str] = []
        for update in self.get_updates():
            try:
                if "elapsed" in update:
                    continue
            except Exception:
                pass
            my_updates.append(update)
        self.assertEqual(my_updates, updates, msg)

    def add_update(self, upd: tuple[str, Any] | str) -> None:
        self.protocol_command.updates.append(upd)

    def patch_runprocess(self, *expectations: runprocess.Expect) -> None:
        """
        Patch a fake RunProcess class in, and set the given expectations.
        """
        assert isinstance(self, TestCase)

        self.patch(buildbot_worker.runprocess, 'RunProcess', runprocess.FakeRunProcess)
        cast(
            "runprocess.FakeRunProcess",
            buildbot_worker.runprocess.RunProcess,
        ).expect(*expectations)
        self.runprocess_patched = True

    def patch_getCommand(self, name: str, result: str) -> None:
        """
        Patch utils.getCommand to return RESULT for NAME
        """
        assert isinstance(self, TestCase)

        old_getCommand = utils.getCommand

        def new_getCommand(n: str) -> str:
            if n == name:
                return result
            return old_getCommand(n)

        self.patch(utils, 'getCommand', new_getCommand)

    def clean_environ(self) -> None:
        """
        Temporarily clean out os.environ to { 'PWD' : '.' }
        """
        assert isinstance(self, TestCase)

        self.patch(os, 'environ', {'PWD': '.'})



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import os
from pathlib import PurePosixPath

from twisted.internet import defer
from twisted.python.filepath import FilePath
from twisted.spread import pb
from twisted.trial.unittest import SkipTest

from buildbot.process import properties
from buildbot.test.fake import fakeprotocol
from buildbot.util.twisted import async_to_deferred
from buildbot.worker import Worker

RemoteWorker: type | None = None
try:
    from buildbot_worker.bot import LocalWorker as RemoteWorker
except ImportError:
    pass


class FakeWorker:
    workername = 'test'

    def __init__(self, master):
        self.master = master
        self.conn = fakeprotocol.FakeConnection(self)
        self.info = properties.Properties()
        self.properties = properties.Properties()
        self.defaultProperties = properties.Properties()
        self.workerid = 383

    def acquireLocks(self):
        return True

    def releaseLocks(self):
        pass

    def attached(self, conn):
        self.worker_system = 'posix'
        self.path_module = os.path
        self.path_cls = PurePosixPath
        self.workerid = 1234
        self.worker_basedir = '/wrk'
        return defer.succeed(None)

    def detached(self):
        pass

    def messageReceivedFromWorker(self):
        pass

    def addWorkerForBuilder(self, wfb):
        pass

    def removeWorkerForBuilder(self, wfb):
        pass

    def buildFinished(self, wfb):
        pass

    def canStartBuild(self):
        pass

    def putInQuarantine(self):
        pass

    def resetQuarantine(self):
        pass


@async_to_deferred
async def disconnect_master_side_worker(worker) -> None:
    # Force disconnection because the LocalWorker does not disconnect itself. Note that
    # the worker may have already been disconnected by something else (e.g. if it's not
    # responding). We need to call detached() explicitly because the order in which
    # disconnection subscriptions are invoked is unspecified.
    if worker.conn is not None:
        worker._detached_sub.unsubscribe()
        conn = worker.conn
        await worker.detached()
        conn.loseConnection()
    await worker.waitForCompleteShutdown()


class SeverWorkerConnectionMixin:
    _connection_severed = False
    _severed_deferreds = None

    def disconnect_worker(self) -> defer.Deferred[None]:
        if not self._connection_severed:
            return defer.succeed(None)

        if self._severed_deferreds is not None:
            for d in self._severed_deferreds:
                d.errback(pb.PBConnectionLost('lost connection'))

        self._connection_severed = False

        return defer.succeed(None)

    def sever_connection(self):
        # stubs the worker connection so that it appears that the TCP connection
        # has been severed in a way that no response is ever received, but
        # messages don't fail immediately. All callback will be called when
        # disconnect_worker is called
        self._connection_severed = True

        def register_deferred():
            d = defer.Deferred()

            if self._severed_deferreds is None:
                self._severed_deferreds = []
            self._severed_deferreds.append(d)

            return d

        def remotePrint(message):
            return register_deferred()

        self.worker.conn.remotePrint = remotePrint

        def remoteGetWorkerInfo():
            return register_deferred()

        self.worker.conn.remoteGetWorkerInfo = remoteGetWorkerInfo

        def remoteSetBuilderList(builders):
            return register_deferred()

        self.worker.conn.remoteSetBuilderList = remoteSetBuilderList

        def remoteStartCommand(remoteCommand, builderName, commandId, commandName, args):
            return register_deferred()

        self.worker.conn.remoteStartCommand = remoteStartCommand

        def remoteShutdown():
            return register_deferred()

        self.worker.conn.remoteShutdown = remoteShutdown

        def remoteStartBuild(builderName):
            return register_deferred()

        self.worker.conn.remoteStartBuild = remoteStartBuild

        def remoteInterruptCommand(builderName, commandId, why):
            return register_deferred()

        self.worker.conn.remoteInterruptCommand = remoteInterruptCommand


class WorkerController(SeverWorkerConnectionMixin):
    """
    A controller for a ``Worker``.

    https://glyph.twistedmatrix.com/2015/05/separate-your-fakes-and-your-inspectors.html
    """

    def __init__(self, case, name, build_wait_timeout=600, worker_class=None, **kwargs):
        if worker_class is None:
            worker_class = Worker

        self.case = case
        self.build_wait_timeout = build_wait_timeout
        self.worker = worker_class(name, self, **kwargs)
        self.remote_worker = None

    @async_to_deferred
    async def connect_worker(self):
        if self.remote_worker is not None:
            return
        if RemoteWorker is None:
            raise SkipTest("buildbot-worker package is not installed")
        workdir = FilePath(self.case.mktemp())
        workdir.createDirectory()
        self.remote_worker = RemoteWorker(self.worker.name, workdir.path, False)
        self.remote_worker.setServiceParent(self.worker)

    @async_to_deferred
    async def disconnect_worker(self):
        await super().disconnect_worker()
        if self.remote_worker is None:
            return

        worker = self.remote_worker
        self.remote_worker = None
        await worker.disownServiceParent()
        await disconnect_master_side_worker(self.worker)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os
import shutil
import socket

from twisted.cred import checkers
from twisted.cred import portal
from twisted.internet import defer
from twisted.internet import reactor
from twisted.spread import pb
from twisted.trial import unittest
from zope.interface import implementer

from buildbot_worker import bot
from buildbot_worker.test.util import misc

try:
    from unittest.mock import Mock
except ImportError:
    from unittest.mock import Mock


# I don't see any simple way to test the PB equipment without actually setting
# up a TCP connection.  This just tests that the PB code will connect and can
# execute a basic ping.  The rest is done without TCP (or PB) in other
# test modules.


class MasterPerspective(pb.Avatar):
    def __init__(self, on_keepalive=None):
        self.on_keepalive = on_keepalive

    def perspective_keepalive(self):
        if self.on_keepalive:
            on_keepalive = self.on_keepalive
            self.on_keepalive = None
            on_keepalive()


@implementer(portal.IRealm)
class MasterRealm:
    def __init__(self, perspective, on_attachment):
        self.perspective = perspective
        self.on_attachment = on_attachment

    @defer.inlineCallbacks
    def requestAvatar(self, avatarId, mind, *interfaces):
        assert pb.IPerspective in interfaces
        self.mind = mind
        self.perspective.mind = mind
        if self.on_attachment:
            yield self.on_attachment(mind)

        return pb.IPerspective, self.perspective, lambda: None

    def shutdown(self):
        return self.mind.broker.transport.loseConnection()


class TestWorker(misc.PatcherMixin, unittest.TestCase):
    def setUp(self):
        self.realm = None
        self.worker = None
        self.listeningport = None

        self.basedir = os.path.abspath("basedir")
        if os.path.exists(self.basedir):
            shutil.rmtree(self.basedir)
        os.makedirs(self.basedir)

    @defer.inlineCallbacks
    def tearDown(self):
        if self.realm:
            yield self.realm.shutdown()
        if self.listeningport:
            yield self.listeningport.stopListening()
        if os.path.exists(self.basedir):
            shutil.rmtree(self.basedir)

    def start_master(self, perspective, on_attachment=None):
        self.realm = MasterRealm(perspective, on_attachment)
        p = portal.Portal(self.realm)
        p.registerChecker(checkers.InMemoryUsernamePasswordDatabaseDontUse(testy=b"westy"))
        self.listeningport = reactor.listenTCP(0, pb.PBServerFactory(p), interface='127.0.0.1')
        # return the dynamically allocated port number
        return self.listeningport.getHost().port

    def test_constructor_minimal(self):
        # only required arguments
        bot.Worker('mstr', 9010, 'me', 'pwd', '/s', 10, protocol='pb')

    def test_constructor_083_tac(self):
        """invocation as made from default 0.8.3 tac files"""
        bot.Worker('mstr', 9010, 'me', 'pwd', '/s', 10, umask=0o123, protocol='pb', maxdelay=10)

    def test_constructor_091_tac(self):
        # invocation as made from default 0.9.1 tac files
        bot.Worker(
            None,
            None,
            'me',
            'pwd',
            '/s',
            10,
            connection_string="tcp:host=localhost:port=9010",
            umask=0o123,
            protocol='pb',
            maxdelay=10,
        )

    def test_constructor_invalid_both_styles(self):
        """Can't instantiate with both host/port and connection string."""
        # assertRaises as a context manager appears in Python 2.7
        self.assertRaises(
            AssertionError,
            bot.Worker,
            'mstr',
            9010,
            'me',
            'pwd',
            '/s',
            10,
            connection_string="tcp:anything",
        )

    def test_constructor_invalid_both_styles_partial(self):
        # assertRaises as a context manager appears in Python 2.7
        self.assertRaises(
            AssertionError,
            bot.Worker,
            'mstr',
            None,
            'me',
            'pwd',
            '/s',
            10,
            connection_string="tcp:anything",
        )

    def test_constructor_invalid_both_styles_partial2(self):
        """Can't instantiate with both host/port and connection string."""
        # assertRaises as a context manager appears in Python 2.7
        self.assertRaises(
            AssertionError,
            bot.Worker,
            None,
            9010,
            None,
            'me',
            'pwd',
            '/s',
            10,
            connection_string="tcp:anything",
        )

    def test_constructor_full(self):
        # invocation with all args
        bot.Worker(
            'mstr',
            9010,
            'me',
            'pwd',
            '/s',
            10,
            umask=0o123,
            maxdelay=10,
            keepaliveTimeout=10,
            unicode_encoding='utf8',
            protocol='pb',
            allow_shutdown=True,
        )

    def test_worker_print(self):
        d = defer.Deferred()

        # set up to call print when we are attached, and chain the results onto
        # the deferred for the whole test
        def call_print(mind):
            print_d = mind.callRemote("print", "Hi, worker.")
            print_d.addCallbacks(d.callback, d.errback)

        # start up the master and worker
        persp = MasterPerspective()
        port = self.start_master(persp, on_attachment=call_print)
        self.worker = bot.Worker(
            "127.0.0.1",
            port,
            "testy",
            "westy",
            self.basedir,
            keepalive=0,
            umask=0o22,
            protocol='pb',
        )
        self.worker.startService()
        self.addCleanup(self.worker.stopService)

        # and wait for the result of the print
        return d

    def test_recordHostname_uname(self):
        self.patch_os_uname(lambda: [0, 'test-hostname.domain.com'])

        self.worker = bot.Worker(
            "127.0.0.1",
            9999,
            "testy",
            "westy",
            self.basedir,
            keepalive=0,
            umask=0o22,
            protocol='pb',
        )
        self.worker.recordHostname(self.basedir)
        with open(os.path.join(self.basedir, "twistd.hostname")) as f:
            twistdHostname = f.read().strip()
        self.assertEqual(twistdHostname, 'test-hostname.domain.com')

    def test_recordHostname_getfqdn(self):
        def missing():
            raise AttributeError

        self.patch_os_uname(missing)
        self.patch(socket, "getfqdn", lambda: 'test-hostname.domain.com')

        self.worker = bot.Worker(
            "127.0.0.1",
            9999,
            "testy",
            "westy",
            self.basedir,
            keepalive=0,
            umask=0o22,
            protocol='pb',
        )
        self.worker.recordHostname(self.basedir)
        with open(os.path.join(self.basedir, "twistd.hostname")) as f:
            twistdHostname = f.read().strip()
        self.assertEqual(twistdHostname, 'test-hostname.domain.com')

    def test_worker_graceful_shutdown(self):
        """Test that running the build worker's gracefulShutdown method results
        in a call to the master's shutdown method"""
        d = defer.Deferred()

        fakepersp = Mock()
        called = []

        def fakeCallRemote(*args):
            called.append(args)
            d1 = defer.succeed(None)
            return d1

        fakepersp.callRemote = fakeCallRemote

        # set up to call shutdown when we are attached, and chain the results onto
        # the deferred for the whole test
        def call_shutdown(mind):
            self.worker.bf.perspective = fakepersp
            shutdown_d = self.worker.gracefulShutdown()
            shutdown_d.addCallbacks(d.callback, d.errback)

        persp = MasterPerspective()
        port = self.start_master(persp, on_attachment=call_shutdown)

        self.worker = bot.Worker(
            "127.0.0.1",
            port,
            "testy",
            "westy",
            self.basedir,
            keepalive=0,
            umask=0o22,
            protocol='pb',
        )

        self.worker.startService()
        self.addCleanup(self.worker.stopService)

        def check(ign):
            self.assertEqual(called, [('shutdown',)])

        d.addCallback(check)

        return d

    def test_worker_shutdown(self):
        """Test watching an existing shutdown_file results in gracefulShutdown
        being called."""

        worker = bot.Worker(
            "127.0.0.1",
            1234,
            "testy",
            "westy",
            self.basedir,
            keepalive=0,
            umask=0o22,
            protocol='pb',
            allow_shutdown='file',
        )

        # Mock out gracefulShutdown
        worker.gracefulShutdown = Mock()

        # Mock out os.path methods
        exists = Mock()
        mtime = Mock()

        self.patch(os.path, 'exists', exists)
        self.patch(os.path, 'getmtime', mtime)

        # Pretend that the shutdown file doesn't exist
        mtime.return_value = 0
        exists.return_value = False

        worker._checkShutdownFile()

        # We shouldn't have called gracefulShutdown
        self.assertEqual(worker.gracefulShutdown.call_count, 0)

        # Pretend that the file exists now, with an mtime of 2
        exists.return_value = True
        mtime.return_value = 2
        worker._checkShutdownFile()

        # Now we should have changed gracefulShutdown
        self.assertEqual(worker.gracefulShutdown.call_count, 1)

        # Bump the mtime again, and make sure we call shutdown again
        mtime.return_value = 3
        worker._checkShutdownFile()
        self.assertEqual(worker.gracefulShutdown.call_count, 2)

        # Try again, we shouldn't call shutdown another time
        worker._checkShutdownFile()
        self.assertEqual(worker.gracefulShutdown.call_count, 2)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.trial import unittest

from buildbot.db import codebase_branches
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util.twisted import async_to_deferred


class Tests(TestReactorMixin, unittest.TestCase):
    @async_to_deferred
    async def setUp(self) -> None:  # type: ignore[override]
        self.setup_test_reactor()
        self.master = await fakemaster.make_master(self, wantDb=True)
        await self.master.db.insert_test_data([
            fakedb.Project(id=7, name='fake_project7'),
            fakedb.Codebase(id=13, projectid=7, name='codebase1'),
            fakedb.CodebaseCommit(id=106, codebaseid=13),
            fakedb.CodebaseCommit(id=107, codebaseid=13),
            fakedb.CodebaseBranch(
                id=200, codebaseid=13, name='branch1', commitid=106, last_timestamp=123456700
            ),
            fakedb.CodebaseBranch(
                id=201, codebaseid=13, name='branch2', commitid=106, last_timestamp=123456701
            ),
            fakedb.CodebaseBranch(
                id=202, codebaseid=13, name='branch3', commitid=106, last_timestamp=123456702
            ),
        ])

    @async_to_deferred
    async def test_get_branch_exists_id(self) -> None:
        dbdict = await self.master.db.codebase_branches.get_branch(id=200)
        self.assertEqual(
            dbdict,
            codebase_branches.CodebaseBranchModel(
                id=200,
                codebaseid=13,
                name='branch1',
                commitid=106,
                last_timestamp=123456700,
            ),
        )

    @async_to_deferred
    async def test_get_branch_exists_name(self) -> None:
        dbdict = await self.master.db.codebase_branches.get_branch_by_name(
            codebaseid=13, name='branch1'
        )
        self.assertEqual(
            dbdict,
            codebase_branches.CodebaseBranchModel(
                id=200,
                codebaseid=13,
                name='branch1',
                commitid=106,
                last_timestamp=123456700,
            ),
        )

    @async_to_deferred
    async def test_get_branch_does_not_exist_name(self) -> None:
        dbdict = await self.master.db.codebase_branches.get_branch_by_name(
            codebaseid=13, name='branch_not_exists'
        )
        self.assertIsNone(dbdict)

    @async_to_deferred
    async def test_get_branch_does_not_exist_id(self) -> None:
        dbdict = await self.master.db.codebase_branches.get_branch(id=234)
        self.assertIsNone(dbdict)

    @async_to_deferred
    async def test_get_branches(self) -> None:
        branches = await self.master.db.codebase_branches.get_branches(codebaseid=13)
        branches = sorted(branches, key=lambda x: x.id)
        self.assertEqual(
            branches,
            [
                codebase_branches.CodebaseBranchModel(
                    id=200,
                    codebaseid=13,
                    name='branch1',
                    commitid=106,
                    last_timestamp=123456700,
                ),
                codebase_branches.CodebaseBranchModel(
                    id=201,
                    codebaseid=13,
                    name='branch2',
                    commitid=106,
                    last_timestamp=123456701,
                ),
                codebase_branches.CodebaseBranchModel(
                    id=202,
                    codebaseid=13,
                    name='branch3',
                    commitid=106,
                    last_timestamp=123456702,
                ),
            ],
        )

    @async_to_deferred
    async def test_update_branch(self) -> None:
        await self.master.db.codebase_branches.update_branch(
            codebaseid=13,
            name='branch1',
            commitid=107,
            last_timestamp=123456789,
        )
        dbdict = await self.master.db.codebase_branches.get_branch(id=200)
        self.assertEqual(
            dbdict,
            codebase_branches.CodebaseBranchModel(
                id=200,
                codebaseid=13,
                name='branch1',
                commitid=107,
                last_timestamp=123456789,
            ),
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.internet import reactor
from twisted.trial import unittest

from buildbot.clients import sendchange as sendchange_client
from buildbot.scripts import sendchange
from buildbot.test.util import misc


class TestSendChange(misc.StdoutAssertionsMixin, unittest.TestCase):
    class FakeSender:
        def __init__(self, testcase, master, auth, encoding=None):
            self.master = master
            self.auth = auth
            self.encoding = encoding
            self.testcase = testcase

        def send(self, branch, revision, comments, files, **kwargs):
            kwargs['branch'] = branch
            kwargs['revision'] = revision
            kwargs['comments'] = comments
            kwargs['files'] = files
            self.send_kwargs = kwargs
            d = defer.Deferred()
            if self.testcase.fail:
                reactor.callLater(0, d.errback, RuntimeError("oh noes"))
            else:
                reactor.callLater(0, d.callback, None)
            return d

    def setUp(self):
        self.fail = False  # set to true to get Sender.send to fail

        def Sender_constr(*args, **kwargs):
            self.sender = self.FakeSender(self, *args, **kwargs)
            return self.sender

        self.patch(sendchange_client, 'Sender', Sender_constr)

        # undo the effects of @in_reactor
        self.patch(sendchange, 'sendchange', sendchange.sendchange._orig)

        self.setUpStdoutAssertions()

    @defer.inlineCallbacks
    def test_sendchange_config(self):
        rc = yield sendchange.sendchange({
            "encoding": 'utf16',
            "who": 'me',
            "auth": ['a', 'b'],
            "master": 'm',
            "branch": 'br',
            "category": 'cat',
            "revision": 'rr',
            "properties": {'a': 'b'},
            "repository": 'rep',
            "project": 'prj',
            "vc": 'git',
            "revlink": 'rl',
            "when": 1234.0,
            "comments": 'comm',
            "files": ('a', 'b'),
            "codebase": 'cb',
        })

        self.assertEqual(
            (
                self.sender.master,
                self.sender.auth,
                self.sender.encoding,
                self.sender.send_kwargs,
                self.getStdout(),
                rc,
            ),
            (
                'm',
                ['a', 'b'],
                'utf16',
                {
                    'branch': 'br',
                    'category': 'cat',
                    'codebase': 'cb',
                    'comments': 'comm',
                    'files': ('a', 'b'),
                    'project': 'prj',
                    'properties': {'a': 'b'},
                    'repository': 'rep',
                    'revision': 'rr',
                    'revlink': 'rl',
                    'when': 1234.0,
                    'who': 'me',
                    'vc': 'git',
                },
                'change sent successfully',
                0,
            ),
        )

    @defer.inlineCallbacks
    def test_sendchange_config_no_codebase(self):
        rc = yield sendchange.sendchange({
            "encoding": 'utf16',
            "who": 'me',
            "auth": ['a', 'b'],
            "master": 'm',
            "branch": 'br',
            "category": 'cat',
            "revision": 'rr',
            "properties": {'a': 'b'},
            "repository": 'rep',
            "project": 'prj',
            "vc": 'git',
            "revlink": 'rl',
            "when": 1234.0,
            "comments": 'comm',
            "files": ('a', 'b'),
        })

        self.assertEqual(
            (
                self.sender.master,
                self.sender.auth,
                self.sender.encoding,
                self.sender.send_kwargs,
                self.getStdout(),
                rc,
            ),
            (
                'm',
                ['a', 'b'],
                'utf16',
                {
                    'branch': 'br',
                    'category': 'cat',
                    'codebase': None,
                    'comments': 'comm',
                    'files': ('a', 'b'),
                    'project': 'prj',
                    'properties': {'a': 'b'},
                    'repository': 'rep',
                    'revision': 'rr',
                    'revlink': 'rl',
                    'when': 1234.0,
                    'who': 'me',
                    'vc': 'git',
                },
                'change sent successfully',
                0,
            ),
        )

    @defer.inlineCallbacks
    def test_sendchange_fail(self):
        self.fail = True
        rc = yield sendchange.sendchange({})

        self.assertEqual((self.getStdout().split('\n')[0], rc), ('change not sent:', 1))



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os
import tarfile
import tempfile

from twisted.internet import defer
from twisted.python import log

from buildbot_worker.commands.base import Command


class TransferCommand(Command):
    def finished(self, res):
        if self.debug:
            self.log_msg(f'finished: stderr={self.stderr!r}, rc={self.rc!r}')

        # don't use self.sendStatus here, since we may no longer be running
        # if we have been interrupted
        updates = [('rc', self.rc)]
        if self.stderr:
            updates.append(('stderr', self.stderr))
        self.protocol_command.send_update(updates)
        return res

    def interrupt(self):
        if self.debug:
            self.log_msg('interrupted')
        if self.interrupted:
            return
        self.rc = 1
        self.interrupted = True
        # now we wait for the next trip around the loop.  It abandon the file
        # when it sees self.interrupted set.


class WorkerFileUploadCommand(TransferCommand):
    """
    Upload a file from worker to build master
    Arguments:

        - ['path']:      path to read from
        - ['writer']:    RemoteReference to a buildbot_worker.protocols.base.FileWriterProxy object
        - ['maxsize']:   max size (in bytes) of file to write
        - ['blocksize']: max size for each data block
        - ['keepstamp']: whether to preserve file modified and accessed times
    """

    debug = False

    requiredArgs = ['path', 'writer', 'blocksize']

    def setup(self, args):
        self.path = args['path']
        self.writer = args['writer']
        self.remaining = args['maxsize']
        self.blocksize = args['blocksize']
        self.keepstamp = args.get('keepstamp', False)
        self.stderr = None
        self.rc = 0
        self.fp = None

    def start(self):
        if self.debug:
            self.log_msg('WorkerFileUploadCommand started')

        access_time = None
        modified_time = None
        try:
            if self.keepstamp:
                access_time = os.path.getatime(self.path)
                modified_time = os.path.getmtime(self.path)

            self.fp = open(self.path, 'rb')
            if self.debug:
                self.log_msg(f"Opened '{self.path}' for upload")
        except Exception:
            self.fp = None
            self.stderr = f"Cannot open file '{self.path}' for upload"
            self.rc = 1
            if self.debug:
                self.log_msg(f"Cannot open file '{self.path}' for upload")

        self.sendStatus([('header', f"sending {self.path}\n")])

        d = defer.Deferred()
        self._reactor.callLater(0, self._loop, d)

        @defer.inlineCallbacks
        def _close_ok(res):
            if self.fp:
                self.fp.close()
            self.fp = None
            yield self.protocol_command.protocol_update_upload_file_close(self.writer)

            if self.keepstamp:
                yield self.protocol_command.protocol_update_upload_file_utime(
                    self.writer, access_time, modified_time
                )

        def _close_err(f):
            self.rc = 1
            if self.fp:
                self.fp.close()
            self.fp = None
            # call remote's close(), but keep the existing failure
            d1 = self.protocol_command.protocol_update_upload_file_close(self.writer)

            def eb(f2):
                self.log_msg("ignoring error from remote close():")
                log.err(f2)

            d1.addErrback(eb)
            d1.addBoth(lambda _: f)  # always return _loop failure
            return d1

        d.addCallbacks(_close_ok, _close_err)
        d.addBoth(self.finished)
        return d

    def _loop(self, fire_when_done):
        d = defer.maybeDeferred(self._writeBlock)

        def _done(finished):
            if finished:
                fire_when_done.callback(None)
            else:
                self._loop(fire_when_done)

        def _err(why):
            fire_when_done.errback(why)

        d.addCallbacks(_done, _err)
        return None

    def _writeBlock(self):
        """Write a block of data to the remote writer"""

        if self.interrupted or self.fp is None:
            if self.debug:
                self.log_msg('WorkerFileUploadCommand._writeBlock(): end')
            return True

        length = self.blocksize
        if self.remaining is not None and length > self.remaining:
            length = self.remaining

        if length <= 0:
            if self.stderr is None:
                self.stderr = f'Maximum filesize reached, truncating file \'{self.path}\''
                self.rc = 1
            data = ''
        else:
            data = self.fp.read(length)

        if self.debug:
            self.log_msg(
                'WorkerFileUploadCommand._writeBlock(): ' + f'allowed={length} readlen={len(data)}'
            )
        if not data:
            self.log_msg("EOF: callRemote(close)")
            return True

        if self.remaining is not None:
            self.remaining = self.remaining - len(data)
            assert self.remaining >= 0
        d = self.do_protocol_write(data)
        d.addCallback(lambda res: False)
        return d

    def do_protocol_write(self, data):
        return self.protocol_command.protocol_update_upload_file_write(self.writer, data)


class WorkerDirectoryUploadCommand(WorkerFileUploadCommand):
    debug = False
    requiredArgs = ['path', 'writer', 'blocksize']

    def setup(self, args):
        self.path = args['path']
        self.writer = args['writer']
        self.remaining = args['maxsize']
        self.blocksize = args['blocksize']
        self.compress = args['compress']
        self.stderr = None
        self.rc = 0

    def start(self):
        if self.debug:
            self.log_msg('WorkerDirectoryUploadCommand started')

        if self.debug:
            self.log_msg(f"path: {self.path!r}")

        # Create temporary archive
        fd, self.tarname = tempfile.mkstemp(prefix='buildbot-transfer-')
        self.fp = os.fdopen(fd, "rb+")
        if self.compress == 'bz2':
            mode = 'w|bz2'
        elif self.compress == 'gz':
            mode = 'w|gz'
        else:
            mode = 'w'

        with tarfile.open(mode=mode, fileobj=self.fp) as archive:
            try:
                archive.add(self.path, '')
            except OSError as e:
                # if directory does not exist, bail out with an error
                self.stderr = f"Cannot read directory '{self.path}' for upload: {e}"
                self.rc = 1
                archive.close()  # need to close it before self.finished() runs below
                d = defer.succeed(False)
                d.addCallback(self.finished)
                return d

        # Transfer it
        self.fp.seek(0)

        self.sendStatus([('header', f"sending {self.path}\n")])

        d = defer.Deferred()
        self._reactor.callLater(0, self._loop, d)

        def unpack(res):
            d1 = self.protocol_command.protocol_update_upload_directory(self.writer)

            def unpack_err(f):
                self.rc = 1
                return f

            d1.addErrback(unpack_err)
            d1.addCallback(lambda ignored: res)
            return d1

        d.addCallback(unpack)
        d.addBoth(self.finished)
        return d

    def finished(self, res):
        self.fp.close()
        self.fp = None
        os.remove(self.tarname)
        return TransferCommand.finished(self, res)

    def do_protocol_write(self, data):
        return self.protocol_command.protocol_update_upload_directory_write(self.writer, data)


class WorkerFileDownloadCommand(TransferCommand):
    """
    Download a file from master to worker
    Arguments:

        - ['path']: path of the worker-side file to be created
        - ['reader']:    RemoteReference to a buildbot_worker.protocols.base.FileReaderProxy object
        - ['maxsize']:   max size (in bytes) of file to write
        - ['blocksize']: max size for each data block
        - ['mode']:      access mode for the new file
    """

    debug = False
    requiredArgs = ['path', 'reader', 'blocksize']

    def setup(self, args):
        self.path = args['path']
        self.reader = args['reader']
        self.bytes_remaining = args['maxsize']
        self.blocksize = args['blocksize']
        self.mode = args['mode']
        self.stderr = None
        self.rc = 0
        self.fp = None

    def start(self):
        if self.debug:
            self.log_msg('WorkerFileDownloadCommand starting')

        dirname = os.path.dirname(self.path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        try:
            self.fp = open(self.path, 'wb')
            if self.debug:
                self.log_msg(f"Opened '{self.path}' for download")
            if self.mode is not None:
                # note: there is a brief window during which the new file
                # will have the worker's default (umask) mode before we
                # set the new one. Don't use this mode= feature to keep files
                # private: use the worker's umask for that instead. (it
                # is possible to call os.umask() before and after the open()
                # call, but cleaning up from exceptions properly is more of a
                # nuisance that way).
                os.chmod(self.path, self.mode)
        except OSError:
            # TODO: this still needs cleanup
            if self.fp:
                self.fp.close()
            self.fp = None
            self.stderr = f"Cannot open file '{self.path}' for download"
            self.rc = 1
            if self.debug:
                self.log_msg(f"Cannot open file '{self.path}' for download")

        d = defer.Deferred()
        self._reactor.callLater(0, self._loop, d)

        def _close(res):
            # close the file, but pass through any errors from _loop
            d1 = self.protocol_command.protocol_update_read_file_close(self.reader)
            d1.addErrback(log.err, 'while trying to close reader')
            d1.addCallback(lambda ignored: res)
            return d1

        d.addBoth(_close)
        d.addBoth(self.finished)
        return d

    def _loop(self, fire_when_done):
        d = defer.maybeDeferred(self._readBlock)

        def _done(finished):
            if finished:
                fire_when_done.callback(None)
            else:
                self._loop(fire_when_done)

        def _err(why):
            fire_when_done.errback(why)

        d.addCallbacks(_done, _err)
        return None

    @defer.inlineCallbacks
    def _readBlock(self):
        """Read a block of data from the remote reader."""

        if self.interrupted or self.fp is None:
            if self.debug:
                self.log_msg('WorkerFileDownloadCommand._readBlock(): end')
            return True

        length = self.blocksize
        if self.bytes_remaining is not None and length > self.bytes_remaining:
            length = self.bytes_remaining

        if length <= 0:
            if self.stderr is None:
                self.stderr = f"Maximum filesize reached, truncating file '{self.path}'"
                self.rc = 1
            return True
        else:
            data = yield self.protocol_command.protocol_update_read_file(self.reader, length)
            return self._writeData(data)

    def _writeData(self, data):
        if self.debug:
            self.log_msg(f'WorkerFileDownloadCommand._readBlock(): readlen={len(data)}')
        if not data:
            return True

        if self.bytes_remaining is not None:
            self.bytes_remaining = self.bytes_remaining - len(data)
            assert self.bytes_remaining >= 0
        self.fp.write(data)
        return False

    def finished(self, res):
        if self.fp:
            self.fp.close()
        self.fp = None

        return TransferCommand.finished(self, res)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import buildbot_worker.commands.fs
import buildbot_worker.commands.shell
import buildbot_worker.commands.transfer

commandRegistry = {
    # command name : fully qualified factory (callable)
    "shell": buildbot_worker.commands.shell.WorkerShellCommand,
    "uploadFile": buildbot_worker.commands.transfer.WorkerFileUploadCommand,
    "upload_file": buildbot_worker.commands.transfer.WorkerFileUploadCommand,
    "uploadDirectory": buildbot_worker.commands.transfer.WorkerDirectoryUploadCommand,
    "upload_directory": buildbot_worker.commands.transfer.WorkerDirectoryUploadCommand,
    "downloadFile": buildbot_worker.commands.transfer.WorkerFileDownloadCommand,
    "download_file": buildbot_worker.commands.transfer.WorkerFileDownloadCommand,
    "mkdir": buildbot_worker.commands.fs.MakeDirectory,
    "rmdir": buildbot_worker.commands.fs.RemoveDirectory,
    "cpdir": buildbot_worker.commands.fs.CopyDirectory,
    "stat": buildbot_worker.commands.fs.StatFile,
    "glob": buildbot_worker.commands.fs.GlobPath,
    "listdir": buildbot_worker.commands.fs.ListDir,
    "rmfile": buildbot_worker.commands.fs.RemoveFile,
}


def getFactory(command):
    factory = commandRegistry[command]
    return factory


def getAllCommandNames():
    return list(commandRegistry)
